from types import SimpleNamespace

import pytest

from sentry_sdk import capture_exception, serializer
from sentry_sdk.integrations.pure_eval import PureEvalIntegration


@pytest.mark.parametrize("integrations", [[], [PureEvalIntegration()]])
def test_include_local_variables_enabled(sentry_init, capture_events, integrations):
    sentry_init(include_local_variables=True, integrations=integrations)
    events = capture_events()

    def foo():
        namespace = SimpleNamespace()
        q = 1
        w = 2
        e = 3
        r = 4
        t = 5
        y = 6
        u = 7
        i = 8
        o = 9
        p = 10
        a = 11
        s = 12
        str((q, w, e, r, t, y, u, i, o, p, a, s))  # use variables for linter
        namespace.d = {1: 2}
        print(namespace.d[1] / 0)

        # Appearances of variables after the main statement don't affect order
        print(q)
        print(s)
        print(events)

    try:
        foo()
    except Exception:
        capture_exception()

    (event,) = events

    assert all(
        frame["vars"]
        for frame in event["exception"]["values"][0]["stacktrace"]["frames"]
    )

    frame_vars = event["exception"]["values"][0]["stacktrace"]["frames"][-1]["vars"]

    if integrations:
        # Values closest to the exception line appear first
        # Test this order if possible given the Python version and dict order
        expected_keys = [
            "namespace",
            "namespace.d",
            "namespace.d[1]",
            "s",
            "a",
            "p",
            "o",
            "i",
            "u",
            "y",
        ]
        assert list(frame_vars.keys()) == expected_keys
        assert frame_vars["namespace.d"] == {"1": "2"}
        assert frame_vars["namespace.d[1]"] == "2"
    else:
        # Without pure_eval, the variables are unpredictable.
        # In later versions, those at the top appear first and are thus included
        assert frame_vars.keys() <= {
            "namespace",
            "q",
            "w",
            "e",
            "r",
            "t",
            "y",
            "u",
            "i",
            "o",
            "p",
            "a",
            "s",
            "events",
        }
        assert len(frame_vars) == serializer.MAX_DATABAG_BREADTH



from typing import Optional

import pytest

from sentry_sdk.integrations.django.middleware import _wrap_middleware


def _sync_capable_middleware_factory(sync_capable: "Optional[bool]") -> type:
    """Create a middleware class with a sync_capable attribute set to the value passed to the factory.
    If the factory is called with None, the middleware class will not have a sync_capable attribute.
    """
    sc = sync_capable  # rename so we can set sync_capable in the class

    class TestMiddleware:
        nonlocal sc
        if sc is not None:
            sync_capable = sc

    return TestMiddleware


@pytest.mark.parametrize(
    ("middleware", "sync_capable"),
    (
        (_sync_capable_middleware_factory(True), True),
        (_sync_capable_middleware_factory(False), False),
        (_sync_capable_middleware_factory(None), True),
    ),
)
def test_wrap_middleware_sync_capable_attribute(middleware, sync_capable):
    wrapped_middleware = _wrap_middleware(middleware, "test_middleware")

    assert wrapped_middleware.sync_capable is sync_capable



"""
These tests exist to verify that Scope.continue_trace() correctly propagates the
sample_rand value onto the transaction's baggage.

We check both the case where there is an incoming sample_rand, as well as the case
where we need to compute it because it is missing.
"""

from unittest import mock
from unittest.mock import Mock

import sentry_sdk


def test_continue_trace_with_sample_rand():
    """
    Test that an incoming sample_rand is propagated onto the transaction's baggage.
    """
    headers = {
        "sentry-trace": "00000000000000000000000000000000-0000000000000000-0",
        "baggage": "sentry-sample_rand=0.1,sentry-sample_rate=0.5",
    }

    transaction = sentry_sdk.continue_trace(headers)
    assert transaction.get_baggage().sentry_items["sample_rand"] == "0.1"


def test_continue_trace_with_sample_rand_span_streaming(sentry_init):
    """
    Test that an incoming sample_rand is propagated onto the segment's baggage.
    """
    sentry_init(traces_sample_rate=1.0, trace_lifecycle="stream")

    headers = {
        "sentry-trace": "00000000000000000000000000000000-0000000000000000-1",
        "baggage": "sentry-sample_rand=0.100000,sentry-sample_rate=0.5",
    }

    sentry_sdk.traces.continue_trace(headers)
    with sentry_sdk.traces.start_span(name="span") as segment:
        assert segment._get_baggage().sentry_items["sample_rand"] == "0.100000"


def test_continue_trace_missing_sample_rand():
    """
    Test that a missing sample_rand is filled in onto the transaction's baggage.
    """

    headers = {
        "sentry-trace": "00000000000000000000000000000000-0000000000000000",
        "baggage": "sentry-placeholder=asdf",
    }

    with mock.patch(
        "sentry_sdk.tracing_utils.Random.randrange", Mock(return_value=500000)
    ):
        transaction = sentry_sdk.continue_trace(headers)

    assert transaction.get_baggage().sentry_items["sample_rand"] == "0.500000"


def test_continue_trace_missing_sample_rand_span_streaming(sentry_init):
    """
    Test that a missing sample_rand is filled in onto the segment's baggage.
    """

    sentry_init(traces_sample_rate=1.0, trace_lifecycle="stream")

    headers = {
        "sentry-trace": "00000000000000000000000000000000-0000000000000000",
        "baggage": "sentry-placeholder=asdf",
    }

    with mock.patch(
        "sentry_sdk.tracing_utils.Random.randrange", Mock(return_value=500000)
    ):
        sentry_sdk.traces.continue_trace(headers)
        with sentry_sdk.traces.start_span(name="span") as span:
            assert span._get_baggage().sentry_items["sample_rand"] == "0.500000"



import sys
import warnings

# This is used in _capture_internal_warnings. We need to run this at import
# time because that's where many deprecation warnings might get thrown.
#
# This lives in tests/__init__.py because apparently even tests/conftest.py
# gets loaded too late.
assert "sentry_sdk" not in sys.modules

_warning_recorder_mgr = warnings.catch_warnings(record=True)
_warning_recorder = _warning_recorder_mgr.__enter__()



"""
Unit tests for the Sentry MCP integration with FastMCP.

This test suite verifies that Sentry's MCPIntegration properly instruments
both FastMCP implementations:
- mcp.server.fastmcp.FastMCP (FastMCP from the mcp package)
- fastmcp.FastMCP (standalone fastmcp package)

Tests focus on verifying Sentry integration behavior:
- Integration doesn't break FastMCP functionality
- Span creation when tools/prompts/resources are called through MCP protocol
- Span data accuracy (operation, name, origin, etc.)
- Error capture and instrumentation
- PII and include_prompts flag behavior
- Request context data extraction
- Transport detection (stdio, http, sse)

All tests invoke tools/prompts/resources through the MCP Server's low-level
request handlers (via CallToolRequest, GetPromptRequest, ReadResourceRequest)
to properly trigger Sentry instrumentation and span creation. This ensures
accurate testing of the integration's behavior in real MCP Server scenarios.
"""

import asyncio
import json
from unittest import mock

import anyio
import pytest

import sentry_sdk

try:
    from unittest.mock import AsyncMock
except ImportError:

    class AsyncMock(mock.MagicMock):
        async def __call__(self, *args, **kwargs):
            return super(AsyncMock, self).__call__(*args, **kwargs)


from mcp.server.sse import SseServerTransport
from mcp.server.streamable_http_manager import StreamableHTTPSessionManager

from sentry_sdk import start_transaction
from sentry_sdk.consts import OP, SPANDATA
from sentry_sdk.integrations.mcp import MCPIntegration

try:
    from fastmcp.prompts import Message
except ImportError:
    Message = None


from starlette.applications import Starlette
from starlette.responses import Response
from starlette.routing import Mount, Route

# Try to import both FastMCP implementations
try:
    from mcp.server.fastmcp import FastMCP as MCPFastMCP

    HAS_MCP_FASTMCP = True
except ImportError:
    HAS_MCP_FASTMCP = False
    MCPFastMCP = None

try:
    from fastmcp import FastMCP as StandaloneFastMCP

    HAS_STANDALONE_FASTMCP = True
except ImportError:
    HAS_STANDALONE_FASTMCP = False
    StandaloneFastMCP = None

# Try to import request_ctx for context testing
try:
    from mcp.server.lowlevel.server import request_ctx
except ImportError:
    request_ctx = None

# Try to import MCP types for helper functions
try:
    from mcp.types import CallToolRequest, GetPromptRequest, ReadResourceRequest
except ImportError:
    # If mcp.types not available, tests will be skipped anyway
    CallToolRequest = None
    GetPromptRequest = None
    ReadResourceRequest = None

try:
    from fastmcp import __version__ as FASTMCP_VERSION
except ImportError:
    FASTMCP_VERSION = None

# Collect available FastMCP implementations for parametrization
fastmcp_implementations = []
fastmcp_ids = []

if HAS_MCP_FASTMCP:
    fastmcp_implementations.append(MCPFastMCP)
    fastmcp_ids.append("mcp.server.fastmcp")

if HAS_STANDALONE_FASTMCP:
    fastmcp_implementations.append(StandaloneFastMCP)
    fastmcp_ids.append("fastmcp")


# Helper functions to call tools through MCP Server protocol
def call_tool_through_mcp(mcp_instance, tool_name, arguments):
    """
    Call a tool through MCP Server's low-level handler.
    This properly triggers Sentry instrumentation.

    Args:
        mcp_instance: The FastMCP instance
        tool_name: Name of the tool to call
        arguments: Dictionary of arguments to pass to the tool

    Returns:
        The tool result normalized to {"result": value} format
    """
    handler = mcp_instance._mcp_server.request_handlers[CallToolRequest]
    request = CallToolRequest(
        method="tools/call", params={"name": tool_name, "arguments": arguments}
    )

    result = asyncio.run(handler(request))

    if hasattr(result, "root"):
        result = result.root
    if hasattr(result, "structuredContent") and result.structuredContent:
        result = result.structuredContent
    elif hasattr(result, "content"):
        if result.content:
            text = result.content[0].text
            try:
                result = json.loads(text)
            except (json.JSONDecodeError, TypeError):
                result = text
        else:
            # Empty content means None return
            result = None

    # Normalize return value to consistent format
    # If already a dict, return as-is (tool functions return dicts directly)
    if isinstance(result, dict):
        return result

    # Handle string "None" or "null" as actual None
    if isinstance(result, str) and result in ("None", "null"):
        result = None

    # Wrap primitive values (int, str, bool, None) in dict format for consistency
    return {"result": result}


async def call_tool_through_mcp_async(mcp_instance, tool_name, arguments):
    """Async version of call_tool_through_mcp."""
    handler = mcp_instance._mcp_server.request_handlers[CallToolRequest]
    request = CallToolRequest(
        method="tools/call", params={"name": tool_name, "arguments": arguments}
    )

    result = await handler(request)

    if hasattr(result, "root"):
        result = result.root
    if hasattr(result, "structuredContent") and result.structuredContent:
        result = result.structuredContent
    elif hasattr(result, "content"):
        if result.content:
            text = result.content[0].text
            try:
                result = json.loads(text)
            except (json.JSONDecodeError, TypeError):
                result = text
        else:
            # Empty content means None return
            result = None

    # Normalize return value to consistent format
    # If already a dict, return as-is (tool functions return dicts directly)
    if isinstance(result, dict):
        return result

    # Handle string "None" or "null" as actual None
    if isinstance(result, str) and result in ("None", "null"):
        result = None

    # Wrap primitive values (int, str, bool, None) in dict format for consistency
    return {"result": result}


def call_prompt_through_mcp(mcp_instance, prompt_name, arguments=None):
    """Call a prompt through MCP Server's low-level handler."""
    handler = mcp_instance._mcp_server.request_handlers[GetPromptRequest]
    request = GetPromptRequest(
        method="prompts/get", params={"name": prompt_name, "arguments": arguments or {}}
    )

    result = asyncio.run(handler(request))
    if hasattr(result, "root"):
        result = result.root
    return result


async def call_prompt_through_mcp_async(mcp_instance, prompt_name, arguments=None):
    """Async version of call_prompt_through_mcp."""
    handler = mcp_instance._mcp_server.request_handlers[GetPromptRequest]
    request = GetPromptRequest(
        method="prompts/get", params={"name": prompt_name, "arguments": arguments or {}}
    )

    result = await handler(request)
    if hasattr(result, "root"):
        result = result.root
    return result


def call_resource_through_mcp(mcp_instance, uri):
    """Call a resource through MCP Server's low-level handler."""
    handler = mcp_instance._mcp_server.request_handlers[ReadResourceRequest]
    request = ReadResourceRequest(method="resources/read", params={"uri": str(uri)})

    result = asyncio.run(handler(request))
    if hasattr(result, "root"):
        result = result.root
    return result


async def call_resource_through_mcp_async(mcp_instance, uri):
    """Async version of call_resource_through_mcp."""
    handler = mcp_instance._mcp_server.request_handlers[ReadResourceRequest]
    request = ReadResourceRequest(method="resources/read", params={"uri": str(uri)})

    result = await handler(request)
    if hasattr(result, "root"):
        result = result.root
    return result


# Skip all tests if neither implementation is available
pytestmark = pytest.mark.skipif(
    not (HAS_MCP_FASTMCP or HAS_STANDALONE_FASTMCP),
    reason="Neither mcp.fastmcp nor standalone fastmcp is installed",
)


@pytest.fixture(autouse=True)
def reset_request_ctx():
    """Reset request context before and after each test"""
    if request_ctx is not None:
        try:
            if request_ctx.get() is not None:
                request_ctx.set(None)
        except LookupError:
            pass

    yield

    if request_ctx is not None:
        try:
            request_ctx.set(None)
        except LookupError:
            pass


# =============================================================================
# Tool Handler Tests - Verifying Sentry Integration
# =============================================================================


@pytest.mark.asyncio
@pytest.mark.parametrize("FastMCP", fastmcp_implementations, ids=fastmcp_ids)
@pytest.mark.parametrize(
    "send_default_pii, include_prompts",
    [(True, True), (True, False), (False, True), (False, False)],
)
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_fastmcp_tool_sync(
    sentry_init,
    capture_events,
    capture_items,
    FastMCP,
    send_default_pii,
    include_prompts,
    stdio,
    span_streaming,
):
    """Test that FastMCP synchronous tool handlers create proper spans"""
    sentry_init(
        integrations=[MCPIntegration(include_prompts=include_prompts)],
        traces_sample_rate=1.0,
        send_default_pii=send_default_pii,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    mcp = FastMCP("Test Server")

    @mcp.tool()
    def add_numbers(a: int, b: int) -> dict:
        """Add two numbers together"""
        return {"result": a + b, "operation": "addition"}

    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="custom parent"):
            # Call through MCP protocol to trigger instrumentation
            result = await stdio(
                mcp._mcp_server,
                method="tools/call",
                params={
                    "name": "add_numbers",
                    "arguments": {"a": 10, "b": 5},
                },
                request_id="req-123",
            )

        assert json.loads(result.message.root.result["content"][0]["text"]) == {
            "result": 15,
            "operation": "addition",
        }

        sentry_sdk.flush()
        spans = [item.payload for item in items if item.type == "span"]
        assert len(spans) == 2

        # Verify span structure
        span = spans[0]
        assert span["attributes"]["sentry.op"] == OP.MCP_SERVER
        assert span["attributes"]["sentry.origin"] == "auto.ai.mcp"
        assert span["name"] == "tools/call add_numbers"
        assert span["attributes"][SPANDATA.MCP_TOOL_NAME] == "add_numbers"
        assert span["attributes"][SPANDATA.MCP_METHOD_NAME] == "tools/call"
        assert span["attributes"][SPANDATA.MCP_TRANSPORT] == "stdio"
        assert span["attributes"][SPANDATA.MCP_REQUEST_ID] == "req-123"

        # Check PII-sensitive data
        if send_default_pii and include_prompts:
            assert SPANDATA.MCP_TOOL_RESULT_CONTENT in span["attributes"]
        else:
            assert SPANDATA.MCP_TOOL_RESULT_CONTENT not in span["attributes"]
    else:
        events = capture_events()
        with start_transaction(name="fastmcp tx"):
            # Call through MCP protocol to trigger instrumentation
            result = await stdio(
                mcp._mcp_server,
                method="tools/call",
                params={
                    "name": "add_numbers",
                    "arguments": {"a": 10, "b": 5},
                },
                request_id="req-123",
            )

        assert json.loads(result.message.root.result["content"][0]["text"]) == {
            "result": 15,
            "operation": "addition",
        }

        (tx,) = events
        assert tx["type"] == "transaction"
        assert len(tx["spans"]) == 1

        # Verify span structure
        span = tx["spans"][0]
        assert span["op"] == OP.MCP_SERVER
        assert span["origin"] == "auto.ai.mcp"
        assert span["description"] == "tools/call add_numbers"
        assert span["data"][SPANDATA.MCP_TOOL_NAME] == "add_numbers"
        assert span["data"][SPANDATA.MCP_METHOD_NAME] == "tools/call"
        assert span["data"][SPANDATA.MCP_TRANSPORT] == "stdio"
        assert span["data"][SPANDATA.MCP_REQUEST_ID] == "req-123"

        # Check PII-sensitive data
        if send_default_pii and include_prompts:
            assert SPANDATA.MCP_TOOL_RESULT_CONTENT in span["data"]
        else:
            assert SPANDATA.MCP_TOOL_RESULT_CONTENT not in span["data"]


@pytest.mark.parametrize("FastMCP", fastmcp_implementations, ids=fastmcp_ids)
@pytest.mark.asyncio
@pytest.mark.parametrize(
    "send_default_pii, include_prompts",
    [(True, True), (True, False), (False, True), (False, False)],
)
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_fastmcp_tool_async(
    sentry_init,
    capture_events,
    capture_items,
    FastMCP,
    send_default_pii,
    include_prompts,
    json_rpc,
    select_transactions_with_mcp_spans,
    span_streaming,
):
    """Test that FastMCP async tool handlers create proper spans"""
    sentry_init(
        integrations=[MCPIntegration(include_prompts=include_prompts)],
        traces_sample_rate=1.0,
        send_default_pii=send_default_pii,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    mcp = FastMCP("Test Server")

    session_manager = StreamableHTTPSessionManager(
        app=mcp._mcp_server,
        json_response=True,
    )

    app = Starlette(
        routes=[
            Mount("/mcp", app=session_manager.handle_request),
        ],
        lifespan=lambda app: session_manager.run(),
    )

    @mcp.tool()
    async def multiply_numbers(x: int, y: int) -> dict:
        """Multiply two numbers together"""
        return {"result": x * y, "operation": "multiplication"}

    if span_streaming:
        items = capture_items("span")

        session_id, result = json_rpc(
            app,
            method="tools/call",
            params={
                "name": "multiply_numbers",
                "arguments": {"x": 7, "y": 6},
            },
            request_id="req-456",
        )

        assert json.loads(result.json()["result"]["content"][0]["text"]) == {
            "result": 42,
            "operation": "multiplication",
        }

        sentry_sdk.flush()
        spans = [item.payload for item in items if item.type == "span"]
        spans = [
            span
            for span in spans
            if span["attributes"].get("mcp.method.name") == "tools/call"
        ]
        assert len(spans) == 1
        span = spans[0]

        assert span["attributes"]["sentry.op"] == OP.MCP_SERVER
        assert span["attributes"]["sentry.origin"] == "auto.ai.mcp"
        assert span["name"] == "tools/call multiply_numbers"
        assert span["attributes"][SPANDATA.MCP_TOOL_NAME] == "multiply_numbers"
        assert span["attributes"][SPANDATA.MCP_METHOD_NAME] == "tools/call"
        assert span["attributes"][SPANDATA.MCP_TRANSPORT] == "http"
        assert span["attributes"][SPANDATA.MCP_REQUEST_ID] == "req-456"
        assert span["attributes"][SPANDATA.MCP_SESSION_ID] == session_id

        # Check PII-sensitive data
        if send_default_pii and include_prompts:
            assert SPANDATA.MCP_TOOL_RESULT_CONTENT in span["attributes"]
        else:
            assert SPANDATA.MCP_TOOL_RESULT_CONTENT not in span["attributes"]
    else:
        events = capture_events()

        session_id, result = json_rpc(
            app,
            method="tools/call",
            params={
                "name": "multiply_numbers",
                "arguments": {"x": 7, "y": 6},
            },
            request_id="req-456",
        )

        assert json.loads(result.json()["result"]["content"][0]["text"]) == {
            "result": 42,
            "operation": "multiplication",
        }

        transactions = select_transactions_with_mcp_spans(
            events, method_name="tools/call"
        )
        assert len(transactions) == 1
        tx = transactions[0]
        assert len(tx["spans"]) == 1
        span = tx["spans"][0]

        assert span["op"] == OP.MCP_SERVER
        assert span["origin"] == "auto.ai.mcp"
        assert span["description"] == "tools/call multiply_numbers"
        assert span["data"][SPANDATA.MCP_TOOL_NAME] == "multiply_numbers"
        assert span["data"][SPANDATA.MCP_METHOD_NAME] == "tools/call"
        assert span["data"][SPANDATA.MCP_TRANSPORT] == "http"
        assert span["data"][SPANDATA.MCP_REQUEST_ID] == "req-456"
        assert span["data"][SPANDATA.MCP_SESSION_ID] == session_id

        # Check PII-sensitive data
        if send_default_pii and include_prompts:
            assert SPANDATA.MCP_TOOL_RESULT_CONTENT in span["data"]
        else:
            assert SPANDATA.MCP_TOOL_RESULT_CONTENT not in span["data"]


@pytest.mark.asyncio
@pytest.mark.parametrize("FastMCP", fastmcp_implementations, ids=fastmcp_ids)
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_fastmcp_tool_with_error(
    sentry_init,
    capture_events,
    capture_items,
    FastMCP,
    stdio,
    span_streaming,
):
    """Test that FastMCP tool handler errors are captured properly"""
    sentry_init(
        integrations=[MCPIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    mcp = FastMCP("Test Server")

    @mcp.tool()
    def failing_tool(value: int) -> int:
        """A tool that always fails"""
        raise ValueError("Tool execution failed")

    if span_streaming:
        items = capture_items("event", "span")
        with sentry_sdk.traces.start_span(name="custom parent"):
            result = await stdio(
                mcp._mcp_server,
                method="tools/call",
                params={
                    "name": "failing_tool",
                    "arguments": {"value": 42},
                },
                request_id="req-error",
            )

        # If no exception raised, check if result indicates error
        assert result.message.root.result["isError"] is True

        sentry_sdk.flush()
        # Check span was created
        spans = [item.payload for item in items if item.type == "span"]
        tool_spans = [
            s for s in spans if s["attributes"].get("sentry.op") == OP.MCP_SERVER
        ]

        assert len(tool_spans) == 1

        # Check error event was captured
        events = [item.payload for item in items if item.type == "event"]
        error_events = [e for e in events if e.get("level") == "error"]
        assert len(error_events) >= 1
        error_event = error_events[0]
        assert error_event["exception"]["values"][0]["type"] == "ValueError"
        assert error_event["exception"]["values"][0]["value"] == "Tool execution failed"

        # Verify span is marked with error
        assert tool_spans[0]["attributes"][SPANDATA.MCP_TOOL_RESULT_IS_ERROR] is True
    else:
        events = capture_events()
        with start_transaction(name="fastmcp tx"):
            result = await stdio(
                mcp._mcp_server,
                method="tools/call",
                params={
                    "name": "failing_tool",
                    "arguments": {"value": 42},
                },
                request_id="req-error",
            )

        # If no exception raised, check if result indicates error
        assert result.message.root.result["isError"] is True

        # Should have transaction and error events
        assert len(events) >= 1

        # Check span was created
        tx = [e for e in events if e.get("type") == "transaction"][0]
        tool_spans = [s for s in tx["spans"] if s["op"] == OP.MCP_SERVER]

        assert len(tool_spans) == 1

        # Check error event was captured
        error_events = [e for e in events if e.get("level") == "error"]
        assert len(error_events) >= 1
        error_event = error_events[0]
        assert error_event["exception"]["values"][0]["type"] == "ValueError"
        assert error_event["exception"]["values"][0]["value"] == "Tool execution failed"

        # Verify span is marked with error
        assert tool_spans[0]["data"][SPANDATA.MCP_TOOL_RESULT_IS_ERROR] is True


@pytest.mark.asyncio
@pytest.mark.parametrize("FastMCP", fastmcp_implementations, ids=fastmcp_ids)
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_fastmcp_multiple_tools(
    sentry_init,
    capture_events,
    capture_items,
    FastMCP,
    stdio,
    span_streaming,
):
    """Test that multiple FastMCP tool calls create multiple spans"""
    sentry_init(
        integrations=[MCPIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    mcp = FastMCP("Test Server")

    @mcp.tool()
    def tool_one(x: int) -> int:
        """First tool"""
        return x * 2

    @mcp.tool()
    def tool_two(y: int) -> int:
        """Second tool"""
        return y + 10

    @mcp.tool()
    def tool_three(z: int) -> int:
        """Third tool"""
        return z - 5

    if span_streaming:
        items = capture_items("span")
        with sentry_sdk.traces.start_span(name="custom parent"):
            result1 = await stdio(
                mcp._mcp_server,
                method="tools/call",
                params={
                    "name": "tool_one",
                    "arguments": {"x": 5},
                },
                request_id="req-multi",
            )

            result2 = await stdio(
                mcp._mcp_server,
                method="tools/call",
                params={
                    "name": "tool_two",
                    "arguments": {
                        "y": int(result1.message.root.result["content"][0]["text"])
                    },
                },
                request_id="req-multi",
            )

            result3 = await stdio(
                mcp._mcp_server,
                method="tools/call",
                params={
                    "name": "tool_three",
                    "arguments": {
                        "z": int(result2.message.root.result["content"][0]["text"])
                    },
                },
                request_id="req-multi",
            )

        assert result1.message.root.result["content"][0]["text"] == "10"
        assert result2.message.root.result["content"][0]["text"] == "20"
        assert result3.message.root.result["content"][0]["text"] == "15"

        sentry_sdk.flush()
        # Verify three spans were created
        spans = [item.payload for item in items if item.type == "span"]
        tool_spans = [
            s for s in spans if s["attributes"].get("sentry.op") == OP.MCP_SERVER
        ]
        assert len(tool_spans) == 3
        assert tool_spans[0]["attributes"][SPANDATA.MCP_TOOL_NAME] == "tool_one"
        assert tool_spans[1]["attributes"][SPANDATA.MCP_TOOL_NAME] == "tool_two"
        assert tool_spans[2]["attributes"][SPANDATA.MCP_TOOL_NAME] == "tool_three"
    else:
        events = capture_events()
        with start_transaction(name="fastmcp tx"):
            result1 = await stdio(
                mcp._mcp_server,
                method="tools/call",
                params={
                    "name": "tool_one",
                    "arguments": {"x": 5},
                },
                request_id="req-multi",
            )

            result2 = await stdio(
                mcp._mcp_server,
                method="tools/call",
                params={
                    "name": "tool_two",
                    "arguments": {
                        "y": int(result1.message.root.result["content"][0]["text"])
                    },
                },
                request_id="req-multi",
            )

            result3 = await stdio(
                mcp._mcp_server,
                method="tools/call",
                params={
                    "name": "tool_three",
                    "arguments": {
                        "z": int(result2.message.root.result["content"][0]["text"])
                    },
                },
                request_id="req-multi",
            )

        assert result1.message.root.result["content"][0]["text"] == "10"
        assert result2.message.root.result["content"][0]["text"] == "20"
        assert result3.message.root.result["content"][0]["text"] == "15"

        (tx,) = events
        assert tx["type"] == "transaction"

        # Verify three spans were created
        tool_spans = [s for s in tx["spans"] if s["op"] == OP.MCP_SERVER]
        assert len(tool_spans) == 3
        assert tool_spans[0]["data"][SPANDATA.MCP_TOOL_NAME] == "tool_one"
        assert tool_spans[1]["data"][SPANDATA.MCP_TOOL_NAME] == "tool_two"
        assert tool_spans[2]["data"][SPANDATA.MCP_TOOL_NAME] == "tool_three"


@pytest.mark.asyncio
@pytest.mark.parametrize("FastMCP", fastmcp_implementations, ids=fastmcp_ids)
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_fastmcp_tool_with_complex_return(
    sentry_init,
    capture_events,
    capture_items,
    FastMCP,
    stdio,
    span_streaming,
):
    """Test FastMCP tool with complex nested return value"""
    sentry_init(
        integrations=[MCPIntegration(include_prompts=True)],
        traces_sample_rate=1.0,
        send_default_pii=True,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    mcp = FastMCP("Test Server")

    @mcp.tool()
    def get_user_data(user_id: int) -> dict:
        """Get complex user data"""
        return {
            "id": user_id,
            "name": "Alice",
            "nested": {"preferences": {"theme": "dark", "notifications": True}},
            "tags": ["admin", "verified"],
        }

    if span_streaming:
        items = capture_items("span")
        with sentry_sdk.traces.start_span(name="custom parent"):
            result = await stdio(
                mcp._mcp_server,
                method="tools/call",
                params={
                    "name": "get_user_data",
                    "arguments": {"user_id": 123},
                },
                request_id="req-complex",
            )

        assert json.loads(result.message.root.result["content"][0]["text"]) == {
            "id": 123,
            "name": "Alice",
            "nested": {"preferences": {"theme": "dark", "notifications": True}},
            "tags": ["admin", "verified"],
        }

        sentry_sdk.flush()
        # Verify span was created with complex data
        spans = [item.payload for item in items if item.type == "span"]
        tool_spans = [
            s for s in spans if s["attributes"].get("sentry.op") == OP.MCP_SERVER
        ]
        assert len(tool_spans) == 1
        assert tool_spans[0]["attributes"]["sentry.op"] == OP.MCP_SERVER
        assert tool_spans[0]["attributes"][SPANDATA.MCP_TOOL_NAME] == "get_user_data"
        # Complex return value should be captured since include_prompts=True and send_default_pii=True
        assert SPANDATA.MCP_TOOL_RESULT_CONTENT in tool_spans[0]["attributes"]
    else:
        events = capture_events()
        with start_transaction(name="fastmcp tx"):
            result = await stdio(
                mcp._mcp_server,
                method="tools/call",
                params={
                    "name": "get_user_data",
                    "arguments": {"user_id": 123},
                },
                request_id="req-complex",
            )

        assert json.loads(result.message.root.result["content"][0]["text"]) == {
            "id": 123,
            "name": "Alice",
            "nested": {"preferences": {"theme": "dark", "notifications": True}},
            "tags": ["admin", "verified"],
        }

        (tx,) = events
        assert tx["type"] == "transaction"

        # Verify span was created with complex data
        tool_spans = [s for s in tx["spans"] if s["op"] == OP.MCP_SERVER]
        assert len(tool_spans) == 1
        assert tool_spans[0]["op"] == OP.MCP_SERVER
        assert tool_spans[0]["data"][SPANDATA.MCP_TOOL_NAME] == "get_user_data"
        # Complex return value should be captured since include_prompts=True and send_default_pii=True
        assert SPANDATA.MCP_TOOL_RESULT_CONTENT in tool_spans[0]["data"]


# =============================================================================
# Prompt Handler Tests (if supported)
# =============================================================================


@pytest.mark.asyncio
@pytest.mark.parametrize("FastMCP", fastmcp_implementations, ids=fastmcp_ids)
@pytest.mark.parametrize(
    "send_default_pii, include_prompts",
    [(True, True), (False, False)],
)
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_fastmcp_prompt_sync(
    sentry_init,
    capture_events,
    capture_items,
    FastMCP,
    send_default_pii,
    include_prompts,
    stdio,
    span_streaming,
):
    """Test that FastMCP synchronous prompt handlers create proper spans"""
    sentry_init(
        integrations=[MCPIntegration(include_prompts=include_prompts)],
        traces_sample_rate=1.0,
        send_default_pii=send_default_pii,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    mcp = FastMCP("Test Server")

    # Try to register a prompt handler (may not be supported in all versions)
    if hasattr(mcp, "prompt"):

        @mcp.prompt()
        def code_help_prompt(language: str):
            """Get help for a programming language"""
            message = {
                "role": "user",
                "content": {
                    "type": "text",
                    "text": f"Tell me about {language}",
                },
            }

            if FASTMCP_VERSION is not None and FASTMCP_VERSION.startswith("3"):
                message = Message(message)

            return [message]

        if span_streaming:
            items = capture_items("span")
            with sentry_sdk.traces.start_span(name="custom parent"):
                result = await stdio(
                    mcp._mcp_server,
                    method="prompts/get",
                    params={
                        "name": "code_help_prompt",
                        "arguments": {"language": "python"},
                    },
                    request_id="req-prompt",
                )

            assert result.message.root.result["messages"][0]["role"] == "user"
            assert (
                "python"
                in result.message.root.result["messages"][0]["content"]["text"].lower()
            )

            sentry_sdk.flush()
            # Verify prompt span was created
            spans = [item.payload for item in items if item.type == "span"]
            prompt_spans = [
                s for s in spans if s["attributes"].get("sentry.op") == OP.MCP_SERVER
            ]
            assert len(prompt_spans) == 1
            span = prompt_spans[0]
            assert span["attributes"]["sentry.origin"] == "auto.ai.mcp"
            assert span["name"] == "prompts/get code_help_prompt"
            assert span["attributes"][SPANDATA.MCP_PROMPT_NAME] == "code_help_prompt"

            # Check PII-sensitive data
            if send_default_pii and include_prompts:
                assert SPANDATA.MCP_PROMPT_RESULT_MESSAGE_CONTENT in span["attributes"]
            else:
                assert (
                    SPANDATA.MCP_PROMPT_RESULT_MESSAGE_CONTENT not in span["attributes"]
                )
        else:
            events = capture_events()
            with start_transaction(name="fastmcp tx"):
                result = await stdio(
                    mcp._mcp_server,
                    method="prompts/get",
                    params={
                        "name": "code_help_prompt",
                        "arguments": {"language": "python"},
                    },
                    request_id="req-prompt",
                )

            assert result.message.root.result["messages"][0]["role"] == "user"
            assert (
                "python"
                in result.message.root.result["messages"][0]["content"]["text"].lower()
            )

            (tx,) = events
            assert tx["type"] == "transaction"

            # Verify prompt span was created
            prompt_spans = [s for s in tx["spans"] if s["op"] == OP.MCP_SERVER]
            assert len(prompt_spans) == 1
            span = prompt_spans[0]
            assert span["origin"] == "auto.ai.mcp"
            assert span["description"] == "prompts/get code_help_prompt"
            assert span["data"][SPANDATA.MCP_PROMPT_NAME] == "code_help_prompt"

            # Check PII-sensitive data
            if send_default_pii and include_prompts:
                assert SPANDATA.MCP_PROMPT_RESULT_MESSAGE_CONTENT in span["data"]
            else:
                assert SPANDATA.MCP_PROMPT_RESULT_MESSAGE_CONTENT not in span["data"]


@pytest.mark.parametrize("FastMCP", fastmcp_implementations, ids=fastmcp_ids)
@pytest.mark.asyncio
async def test_fastmcp_prompt_async(
    sentry_init,
    capture_events,
    FastMCP,
    json_rpc,
    select_transactions_with_mcp_spans,
):
    """Test that FastMCP async prompt handlers create proper spans"""
    sentry_init(
        integrations=[MCPIntegration()],
        traces_sample_rate=1.0,
    )
    events = capture_events()

    mcp = FastMCP("Test Server")

    session_manager = StreamableHTTPSessionManager(
        app=mcp._mcp_server,
        json_response=True,
    )

    app = Starlette(
        routes=[
            Mount("/mcp", app=session_manager.handle_request),
        ],
        lifespan=lambda app: session_manager.run(),
    )

    # Try to register an async prompt handler
    if hasattr(mcp, "prompt"):

        @mcp.prompt()
        async def async_prompt(topic: str):
            """Get async prompt for a topic"""
            message1 = {
                "role": "user",
                "content": {"type": "text", "text": f"What is {topic}?"},
            }

            message2 = {
                "role": "assistant",
                "content": {
                    "type": "text",
                    "text": "Let me explain that",
                },
            }

            if FASTMCP_VERSION is not None and FASTMCP_VERSION.startswith("3"):
                message1 = Message(message1)
                message2 = Message(message2)

            return [message1, message2]

        _, result = json_rpc(
            app,
            method="prompts/get",
            params={
                "name": "async_prompt",
                "arguments": {"topic": "MCP"},
            },
            request_id="req-async-prompt",
        )

        assert len(result.json()["result"]["messages"]) == 2

        transactions = select_transactions_with_mcp_spans(
            events, method_name="prompts/get"
        )
        assert len(transactions) == 1


# =============================================================================
# Resource Handler Tests (if supported)
# =============================================================================


@pytest.mark.asyncio
@pytest.mark.parametrize("FastMCP", fastmcp_implementations, ids=fastmcp_ids)
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_fastmcp_resource_sync(
    sentry_init,
    capture_events,
    capture_items,
    FastMCP,
    stdio,
    span_streaming,
):
    """Test that FastMCP synchronous resource handlers create proper spans"""
    sentry_init(
        integrations=[MCPIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    mcp = FastMCP("Test Server")

    # Try to register a resource handler
    try:
        if hasattr(mcp, "resource"):

            @mcp.resource("file:///{path}")
            def read_file(path: str):
                """Read a file resource"""
                return "file contents"

            if span_streaming:
                items = capture_items("span")
                with sentry_sdk.traces.start_span(name="custom parent"):
                    try:
                        result = await stdio(
                            mcp._mcp_server,
                            method="resources/read",
                            params={
                                "uri": "file:///test.txt",
                            },
                            request_id="req-resource",
                        )
                    except ValueError as e:
                        # Older FastMCP versions may not support this URI pattern
                        if "Unknown resource" in str(e):
                            pytest.skip(
                                f"Resource URI not supported in this FastMCP version: {e}"
                            )
                        raise

                # Resource content is returned as-is
                assert (
                    "file contents" in result.message.root.result["contents"][0]["text"]
                )

                sentry_sdk.flush()
                # Verify resource span was created
                spans = [item.payload for item in items if item.type == "span"]
                resource_spans = [
                    s
                    for s in spans
                    if s["attributes"].get("sentry.op") == OP.MCP_SERVER
                ]
                assert len(resource_spans) == 1
                span = resource_spans[0]
                assert span["attributes"]["sentry.origin"] == "auto.ai.mcp"
                assert span["name"] == "resources/read file:///test.txt"
                assert span["attributes"][SPANDATA.MCP_RESOURCE_PROTOCOL] == "file"
            else:
                events = capture_events()
                with start_transaction(name="fastmcp tx"):
                    try:
                        result = await stdio(
                            mcp._mcp_server,
                            method="resources/read",
                            params={
                                "uri": "file:///test.txt",
                            },
                            request_id="req-resource",
                        )
                    except ValueError as e:
                        # Older FastMCP versions may not support this URI pattern
                        if "Unknown resource" in str(e):
                            pytest.skip(
                                f"Resource URI not supported in this FastMCP version: {e}"
                            )
                        raise

                # Resource content is returned as-is
                assert (
                    "file contents" in result.message.root.result["contents"][0]["text"]
                )

                (tx,) = events
                assert tx["type"] == "transaction"

                # Verify resource span was created
                resource_spans = [s for s in tx["spans"] if s["op"] == OP.MCP_SERVER]
                assert len(resource_spans) == 1
                span = resource_spans[0]
                assert span["origin"] == "auto.ai.mcp"
                assert span["description"] == "resources/read file:///test.txt"
                assert span["data"][SPANDATA.MCP_RESOURCE_PROTOCOL] == "file"

    except (AttributeError, TypeError):
        # Resource handler not supported in this version
        pytest.skip("Resource handlers not supported in this FastMCP version")


@pytest.mark.parametrize("FastMCP", fastmcp_implementations, ids=fastmcp_ids)
@pytest.mark.asyncio
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_fastmcp_resource_async(
    sentry_init,
    capture_events,
    capture_items,
    FastMCP,
    json_rpc,
    select_transactions_with_mcp_spans,
    span_streaming,
):
    """Test that FastMCP async resource handlers create proper spans"""
    sentry_init(
        integrations=[MCPIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    mcp = FastMCP("Test Server")

    session_manager = StreamableHTTPSessionManager(
        app=mcp._mcp_server,
        json_response=True,
    )

    app = Starlette(
        routes=[
            Mount("/mcp", app=session_manager.handle_request),
        ],
        lifespan=lambda app: session_manager.run(),
    )

    # Try to register an async resource handler
    try:
        if hasattr(mcp, "resource"):
            if span_streaming:
                items = capture_items("span")

                @mcp.resource("https://example.com/{resource}")
                async def read_url(resource: str):
                    """Read a URL resource"""
                    return "resource data"

                _, result = json_rpc(
                    app,
                    method="resources/read",
                    params={
                        "uri": "https://example.com/resource",
                    },
                    request_id="req-async-resource",
                )
                # Older FastMCP versions may not support this URI pattern
                if (
                    "error" in result.json()
                    and "Unknown resource" in result.json()["error"]["message"]
                ):
                    pytest.skip("Resource URI not supported in this FastMCP version.")
                    return

                assert "resource data" in result.json()["result"]["contents"][0]["text"]

                sentry_sdk.flush()
                spans = [item.payload for item in items if item.type == "span"]
                spans = [
                    span
                    for span in spans
                    if span["attributes"].get("mcp.method.name") == "resources/read"
                ]
                assert len(spans) == 1
                span = spans[0]

                assert span["attributes"][SPANDATA.MCP_RESOURCE_PROTOCOL] == "https"
            else:
                events = capture_events()

                @mcp.resource("https://example.com/{resource}")
                async def read_url(resource: str):
                    """Read a URL resource"""
                    return "resource data"

                _, result = json_rpc(
                    app,
                    method="resources/read",
                    params={
                        "uri": "https://example.com/resource",
                    },
                    request_id="req-async-resource",
                )
                # Older FastMCP versions may not support this URI pattern
                if (
                    "error" in result.json()
                    and "Unknown resource" in result.json()["error"]["message"]
                ):
                    pytest.skip("Resource URI not supported in this FastMCP version.")
                    return

                assert "resource data" in result.json()["result"]["contents"][0]["text"]

                transactions = select_transactions_with_mcp_spans(
                    events, method_name="resources/read"
                )
                assert len(transactions) == 1
                tx = transactions[0]
                assert len(tx["spans"]) == 1
                span = tx["spans"][0]

                assert span["data"][SPANDATA.MCP_RESOURCE_PROTOCOL] == "https"
    except (AttributeError, TypeError):
        # Resource handler not supported in this version
        pytest.skip("Resource handlers not supported in this FastMCP version")


# =============================================================================
# Span Origin and Metadata Tests
# =============================================================================


@pytest.mark.asyncio
@pytest.mark.parametrize("FastMCP", fastmcp_implementations, ids=fastmcp_ids)
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_fastmcp_span_origin(
    sentry_init,
    capture_events,
    capture_items,
    FastMCP,
    stdio,
    span_streaming,
):
    """Test that FastMCP span origin is set correctly"""
    sentry_init(
        integrations=[MCPIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    mcp = FastMCP("Test Server")

    @mcp.tool()
    def test_tool(value: int) -> int:
        """Test tool for origin checking"""
        return value * 2

    if span_streaming:
        items = capture_items("span")
        with sentry_sdk.traces.start_span(name="custom parent"):
            await stdio(
                mcp._mcp_server,
                method="tools/call",
                params={
                    "name": "test_tool",
                    "arguments": {"value": 21},
                },
                request_id="req-origin",
            )

        sentry_sdk.flush()

        spans = [item.payload for item in items if item.type == "span"]
        assert spans[-1]["attributes"]["sentry.origin"] == "manual"

        # Verify MCP span has correct origin
        mcp_spans = [
            s for s in spans if s["attributes"].get("sentry.op") == OP.MCP_SERVER
        ]
        assert len(mcp_spans) == 1
        assert mcp_spans[0]["attributes"]["sentry.origin"] == "auto.ai.mcp"
    else:
        events = capture_events()
        with start_transaction(name="fastmcp tx"):
            await stdio(
                mcp._mcp_server,
                method="tools/call",
                params={
                    "name": "test_tool",
                    "arguments": {"value": 21},
                },
                request_id="req-origin",
            )

        (tx,) = events

        assert tx["contexts"]["trace"]["origin"] == "manual"

        # Verify MCP span has correct origin
        mcp_spans = [s for s in tx["spans"] if s["op"] == OP.MCP_SERVER]
        assert len(mcp_spans) == 1
        assert mcp_spans[0]["origin"] == "auto.ai.mcp"


# =============================================================================
# Transport Detection Tests
# =============================================================================


@pytest.mark.asyncio
@pytest.mark.parametrize("FastMCP", fastmcp_implementations, ids=fastmcp_ids)
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_fastmcp_sse_transport(
    sentry_init,
    capture_events,
    capture_items,
    FastMCP,
    json_rpc_sse,
    span_streaming,
):
    """Test that FastMCP correctly detects SSE transport"""
    sentry_init(
        integrations=[MCPIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    mcp = FastMCP("Test Server")
    sse = SseServerTransport("/messages/")

    sse_connection_closed = asyncio.Event()

    async def handle_sse(request):
        async with sse.connect_sse(
            request.scope, request.receive, request._send
        ) as streams:
            async with anyio.create_task_group() as tg:

                async def run_server():
                    await mcp._mcp_server.run(
                        streams[0],
                        streams[1],
                        mcp._mcp_server.create_initialization_options(),
                    )

                tg.start_soon(run_server)

        sse_connection_closed.set()
        return Response()

    app = Starlette(
        routes=[
            Route("/sse", endpoint=handle_sse, methods=["GET"]),
            Mount("/messages/", app=sse.handle_post_message),
        ],
    )

    @mcp.tool()
    def sse_tool(value: str) -> dict:
        """Tool for SSE transport test"""
        return {"message": f"Received: {value}"}

    keep_sse_alive = asyncio.Event()

    if span_streaming:
        items = capture_items("span")

        app_task, _, result = await json_rpc_sse(
            app,
            method="tools/call",
            params={
                "name": "sse_tool",
                "arguments": {"value": "hello"},
            },
            request_id="req-sse",
            keep_sse_alive=keep_sse_alive,
        )

        await sse_connection_closed.wait()
        await app_task

        assert json.loads(result["result"]["content"][0]["text"]) == {
            "message": "Received: hello"
        }

        sentry_sdk.flush()
        # Find MCP spans
        spans = [item.payload for item in items if item.type == "span"]
        mcp_spans = [
            s for s in spans if s["attributes"].get("sentry.op") == OP.MCP_SERVER
        ]

        assert len(mcp_spans) >= 1
        span = mcp_spans[0]
        # Check that SSE transport is detected
        assert span["attributes"].get(SPANDATA.MCP_TRANSPORT) == "sse"
    else:
        events = capture_events()

        app_task, _, result = await json_rpc_sse(
            app,
            method="tools/call",
            params={
                "name": "sse_tool",
                "arguments": {"value": "hello"},
            },
            request_id="req-sse",
            keep_sse_alive=keep_sse_alive,
        )

        await sse_connection_closed.wait()
        await app_task

        assert json.loads(result["result"]["content"][0]["text"]) == {
            "message": "Received: hello"
        }

        transactions = [
            event
            for event in events
            if event["type"] == "transaction" and event["transaction"] == "/sse"
        ]
        assert len(transactions) == 1
        tx = transactions[0]

        # Find MCP spans
        mcp_spans = [s for s in tx["spans"] if s["op"] == OP.MCP_SERVER]

        assert len(mcp_spans) >= 1
        span = mcp_spans[0]
        # Check that SSE transport is detected
        assert span["data"].get(SPANDATA.MCP_TRANSPORT) == "sse"


@pytest.mark.parametrize("FastMCP", fastmcp_implementations, ids=fastmcp_ids)
@pytest.mark.parametrize("span_streaming", [True, False])
def test_fastmcp_http_transport(
    sentry_init,
    capture_events,
    capture_items,
    FastMCP,
    json_rpc,
    select_transactions_with_mcp_spans,
    span_streaming,
):
    """Test that FastMCP correctly detects HTTP transport"""
    sentry_init(
        integrations=[MCPIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    mcp = FastMCP("Test Server")

    session_manager = StreamableHTTPSessionManager(
        app=mcp._mcp_server,
        json_response=True,
    )

    app = Starlette(
        routes=[
            Mount("/mcp", app=session_manager.handle_request),
        ],
        lifespan=lambda app: session_manager.run(),
    )

    @mcp.tool()
    def http_tool(data: str) -> dict:
        """Tool for HTTP transport test"""
        return {"processed": data.upper()}

    if span_streaming:
        items = capture_items("span")

        _, result = json_rpc(
            app,
            method="tools/call",
            params={
                "name": "http_tool",
                "arguments": {"data": "test"},
            },
            request_id="req-http",
        )

        assert json.loads(result.json()["result"]["content"][0]["text"]) == {
            "processed": "TEST"
        }

        sentry_sdk.flush()
        spans = [item.payload for item in items if item.type == "span"]
        spans = [
            span
            for span in spans
            if span["attributes"].get("mcp.method.name") == "tools/call"
        ]
        assert len(spans) == 1
        span = spans[0]

        # Check that HTTP transport is detected
        assert span["attributes"].get(SPANDATA.MCP_TRANSPORT) == "http"
    else:
        events = capture_events()

        _, result = json_rpc(
            app,
            method="tools/call",
            params={
                "name": "http_tool",
                "arguments": {"data": "test"},
            },
            request_id="req-http",
        )

        assert json.loads(result.json()["result"]["content"][0]["text"]) == {
            "processed": "TEST"
        }

        transactions = select_transactions_with_mcp_spans(
            events, method_name="tools/call"
        )
        assert len(transactions) == 1
        tx = transactions[0]
        assert len(tx["spans"]) == 1
        span = tx["spans"][0]

        # Check that HTTP transport is detected
        assert span["data"].get(SPANDATA.MCP_TRANSPORT) == "http"


@pytest.mark.asyncio
@pytest.mark.parametrize("FastMCP", fastmcp_implementations, ids=fastmcp_ids)
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_fastmcp_stdio_transport(
    sentry_init,
    capture_events,
    capture_items,
    FastMCP,
    stdio,
    span_streaming,
):
    """Test that FastMCP correctly detects stdio transport"""
    sentry_init(
        integrations=[MCPIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    mcp = FastMCP("Test Server")

    @mcp.tool()
    def stdio_tool(n: int) -> dict:
        """Tool for stdio transport test"""
        return {"squared": n * n}

    if span_streaming:
        items = capture_items("span")
        with sentry_sdk.traces.start_span(name="custom parent"):
            result = await stdio(
                mcp._mcp_server,
                method="tools/call",
                params={
                    "name": "stdio_tool",
                    "arguments": {"n": 7},
                },
                request_id="req-stdio",
            )

        assert json.loads(result.message.root.result["content"][0]["text"]) == {
            "squared": 49
        }

        sentry_sdk.flush()
        # Find MCP spans
        spans = [item.payload for item in items if item.type == "span"]
        mcp_spans = [
            s for s in spans if s["attributes"].get("sentry.op") == OP.MCP_SERVER
        ]

        assert len(mcp_spans) >= 1
        span = mcp_spans[0]
        # Check that stdio transport is detected

        assert span["attributes"].get(SPANDATA.MCP_TRANSPORT) == "stdio"
    else:
        events = capture_events()
        with start_transaction(name="fastmcp tx"):
            result = await stdio(
                mcp._mcp_server,
                method="tools/call",
                params={
                    "name": "stdio_tool",
                    "arguments": {"n": 7},
                },
                request_id="req-stdio",
            )

        assert json.loads(result.message.root.result["content"][0]["text"]) == {
            "squared": 49
        }

        (tx,) = events

        # Find MCP spans
        mcp_spans = [s for s in tx["spans"] if s["op"] == OP.MCP_SERVER]

        assert len(mcp_spans) >= 1
        span = mcp_spans[0]
        # Check that stdio transport is detected

        assert span["data"].get(SPANDATA.MCP_TRANSPORT) == "stdio"


# =============================================================================
# Integration-specific Tests
# =============================================================================


@pytest.mark.skipif(not HAS_MCP_FASTMCP, reason="mcp.server.fastmcp not installed")
def test_mcp_fastmcp_specific_features(sentry_init, capture_events):
    """Test features specific to mcp.server.fastmcp (from mcp package)"""
    sentry_init(
        integrations=[MCPIntegration()],
        traces_sample_rate=1.0,
    )
    events = capture_events()

    from mcp.server.fastmcp import FastMCP

    mcp = FastMCP("MCP Package Server")

    @mcp.tool()
    def package_specific_tool(x: int) -> int:
        """Tool for mcp.server.fastmcp package"""
        return x + 100

    with start_transaction(name="mcp.server.fastmcp tx"):
        result = call_tool_through_mcp(mcp, "package_specific_tool", {"x": 50})

    assert result["result"] == 150

    (tx,) = events
    assert tx["type"] == "transaction"


@pytest.mark.asyncio
@pytest.mark.skipif(
    not HAS_STANDALONE_FASTMCP, reason="standalone fastmcp not installed"
)
async def test_standalone_fastmcp_specific_features(sentry_init, capture_events, stdio):
    """Test features specific to standalone fastmcp package"""
    sentry_init(
        integrations=[MCPIntegration()],
        traces_sample_rate=1.0,
    )
    events = capture_events()

    from fastmcp import FastMCP

    mcp = FastMCP("Standalone FastMCP Server")

    @mcp.tool()
    def standalone_specific_tool(message: str) -> dict:
        """Tool for standalone fastmcp package"""
        return {"echo": message, "length": len(message)}

    with start_transaction(name="standalone fastmcp tx"):
        result = await stdio(
            mcp._mcp_server,
            method="tools/call",
            params={
                "name": "standalone_specific_tool",
                "arguments": {"message": "Hello FastMCP"},
            },
        )

    assert json.loads(result.message.root.result["content"][0]["text"]) == {
        "echo": "Hello FastMCP",
        "length": 13,
    }

    (tx,) = events
    assert tx["type"] == "transaction"


# =============================================================================
# Edge Cases and Robustness Tests
# =============================================================================


@pytest.mark.asyncio
@pytest.mark.parametrize("FastMCP", fastmcp_implementations, ids=fastmcp_ids)
async def test_fastmcp_tool_with_no_arguments(
    sentry_init, capture_events, FastMCP, stdio
):
    """Test FastMCP tool with no arguments"""
    sentry_init(
        integrations=[MCPIntegration()],
        traces_sample_rate=1.0,
    )
    events = capture_events()

    mcp = FastMCP("Test Server")

    @mcp.tool()
    def no_args_tool() -> str:
        """Tool that takes no arguments"""
        return "success"

    with start_transaction(name="fastmcp tx"):
        result = await stdio(
            mcp._mcp_server,
            method="tools/call",
            params={
                "name": "no_args_tool",
                "arguments": {},
            },
        )

    assert result.message.root.result["content"][0]["text"] == "success"

    (tx,) = events
    assert tx["type"] == "transaction"


@pytest.mark.asyncio
@pytest.mark.parametrize("FastMCP", fastmcp_implementations, ids=fastmcp_ids)
async def test_fastmcp_tool_with_none_return(
    sentry_init, capture_events, FastMCP, stdio
):
    """Test FastMCP tool that returns None"""
    sentry_init(
        integrations=[MCPIntegration()],
        traces_sample_rate=1.0,
    )
    events = capture_events()

    mcp = FastMCP("Test Server")

    @mcp.tool()
    def none_return_tool(action: str) -> None:
        """Tool that returns None"""
        pass

    with start_transaction(name="fastmcp tx"):
        result = await stdio(
            mcp._mcp_server,
            method="tools/call",
            params={
                "name": "none_return_tool",
                "arguments": {"action": "log"},
            },
        )

    if (
        isinstance(mcp, StandaloneFastMCP) and FASTMCP_VERSION is not None
    ) or isinstance(mcp, MCPFastMCP):
        assert len(result.message.root.result["content"]) == 0
    else:
        assert result.message.root.result["content"] == [
            {"type": "text", "text": "None"}
        ]

    (tx,) = events
    assert tx["type"] == "transaction"


@pytest.mark.asyncio
@pytest.mark.parametrize("FastMCP", fastmcp_implementations, ids=fastmcp_ids)
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_fastmcp_mixed_sync_async_tools(
    sentry_init,
    capture_events,
    capture_items,
    FastMCP,
    stdio,
    span_streaming,
):
    """Test mixing sync and async tools in FastMCP"""
    sentry_init(
        integrations=[MCPIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    mcp = FastMCP("Test Server")

    @mcp.tool()
    def sync_add(a: int, b: int) -> int:
        """Sync addition"""
        return a + b

    @mcp.tool()
    async def async_multiply(x: int, y: int) -> int:
        """Async multiplication"""
        return x * y

    if span_streaming:
        items = capture_items("span")
        with sentry_sdk.traces.start_span(name="custom parent"):
            # Use async version for both since we're in an async context
            result1 = await stdio(
                mcp._mcp_server,
                method="tools/call",
                params={
                    "name": "sync_add",
                    "arguments": {"a": 3, "b": 4},
                },
                request_id="req-mixed",
            )
            result2 = await stdio(
                mcp._mcp_server,
                method="tools/call",
                params={
                    "name": "async_multiply",
                    "arguments": {"x": 5, "y": 6},
                },
                request_id="req-mixed",
            )

        assert result1.message.root.result["content"][0]["text"] == "7"
        assert result2.message.root.result["content"][0]["text"] == "30"

        sentry_sdk.flush()
        # Verify both sync and async tool spans were created
        spans = [item.payload for item in items if item.type == "span"]
        mcp_spans = [
            s for s in spans if s["attributes"].get("sentry.op") == OP.MCP_SERVER
        ]
        assert len(mcp_spans) == 2
        assert mcp_spans[0]["attributes"][SPANDATA.MCP_TOOL_NAME] == "sync_add"
        assert mcp_spans[1]["attributes"][SPANDATA.MCP_TOOL_NAME] == "async_multiply"
    else:
        events = capture_events()
        with start_transaction(name="fastmcp tx"):
            # Use async version for both since we're in an async context
            result1 = await stdio(
                mcp._mcp_server,
                method="tools/call",
                params={
                    "name": "sync_add",
                    "arguments": {"a": 3, "b": 4},
                },
                request_id="req-mixed",
            )
            result2 = await stdio(
                mcp._mcp_server,
                method="tools/call",
                params={
                    "name": "async_multiply",
                    "arguments": {"x": 5, "y": 6},
                },
                request_id="req-mixed",
            )

        assert result1.message.root.result["content"][0]["text"] == "7"
        assert result2.message.root.result["content"][0]["text"] == "30"

        (tx,) = events
        assert tx["type"] == "transaction"

        # Verify both sync and async tool spans were created
        mcp_spans = [s for s in tx["spans"] if s["op"] == OP.MCP_SERVER]
        assert len(mcp_spans) == 2
        assert mcp_spans[0]["data"][SPANDATA.MCP_TOOL_NAME] == "sync_add"
        assert mcp_spans[1]["data"][SPANDATA.MCP_TOOL_NAME] == "async_multiply"



"""
Sentry integration for MCP (Model Context Protocol) servers.

This integration instruments MCP servers to create spans for tool, prompt,
and resource handler execution, and captures errors that occur during execution.

Supports the low-level `mcp.server.lowlevel.Server` API.
"""

import inspect
from functools import wraps
from typing import TYPE_CHECKING

import sentry_sdk
from sentry_sdk.ai.utils import _set_span_data_attribute, get_start_span_function
from sentry_sdk.consts import OP, SPANDATA
from sentry_sdk.integrations import DidNotEnable, Integration
from sentry_sdk.scope import should_send_default_pii
from sentry_sdk.traces import StreamedSpan
from sentry_sdk.tracing_utils import has_span_streaming_enabled
from sentry_sdk.utils import nullcontext, package_version, safe_serialize

MCP_PACKAGE_VERSION = package_version("mcp")

try:
    from mcp.server.lowlevel import Server
    from mcp.server.streamable_http import (
        StreamableHTTPServerTransport,
    )

    if MCP_PACKAGE_VERSION and MCP_PACKAGE_VERSION < (2, 0, 0):
        from mcp.server.lowlevel.server import (  # type: ignore[attr-defined]
            request_ctx,
        )
except ImportError:
    raise DidNotEnable("MCP SDK not installed")

try:
    from fastmcp import FastMCP  # type: ignore[import-not-found]
except ImportError:
    FastMCP = None

if MCP_PACKAGE_VERSION and MCP_PACKAGE_VERSION >= (2, 0, 0):
    try:
        from mcp.server.context import (
            ServerRequestContext,
        )
    except ImportError:
        ServerRequestContext = None  # type: ignore[assignment,misc]
else:
    ServerRequestContext = None  # type: ignore[assignment,misc]


if TYPE_CHECKING:
    from typing import Any, Callable, ContextManager, Optional, Tuple, Union

    from starlette.types import Receive, Scope, Send

    from sentry_sdk.traces import StreamedSpan
    from sentry_sdk.tracing import Span


class MCPIntegration(Integration):
    identifier = "mcp"
    origin = "auto.ai.mcp"

    def __init__(self, include_prompts: bool = True) -> None:
        """
        Initialize the MCP integration.

        Args:
            include_prompts: Whether to include prompts (tool results and prompt content)
                             in span data. Requires send_default_pii=True. Default is True.
        """
        self.include_prompts = include_prompts

    @staticmethod
    def setup_once() -> None:
        """
        Patches MCP server classes to instrument handler execution.
        """
        _patch_lowlevel_server()
        _patch_handle_request()

        if FastMCP is not None:
            _patch_fastmcp()


def _get_active_http_scopes(
    ctx: "Optional[Any]" = None,
) -> "Optional[Tuple[Optional[sentry_sdk.Scope], Optional[sentry_sdk.Scope]]]":
    if MCP_PACKAGE_VERSION and MCP_PACKAGE_VERSION < (2, 0, 0):
        if ctx is None:
            try:
                ctx = request_ctx.get()
            except LookupError:
                return None

    if (
        ctx is None
        or not hasattr(ctx, "request")
        or ctx.request is None
        or "state" not in ctx.request.scope
    ):
        return None

    return (
        ctx.request.scope["state"].get("sentry_sdk.isolation_scope"),
        ctx.request.scope["state"].get("sentry_sdk.current_scope"),
    )


def _get_request_context_data(
    ctx: "Optional[Any]" = None,
) -> "tuple[Optional[str], Optional[str], str]":
    """
    Extract request ID, session ID, and MCP transport type from the request context.

    Returns:
        Tuple of (request_id, session_id, mcp_transport).
        - request_id: May be None if not available
        - session_id: May be None if not available
        - mcp_transport: "http", "sse", "stdio"
    """
    request_id: "Optional[str]" = None
    session_id: "Optional[str]" = None
    mcp_transport: str = "stdio"
    if MCP_PACKAGE_VERSION and MCP_PACKAGE_VERSION < (2, 0, 0):
        if ctx is None:
            try:
                ctx = request_ctx.get()
            except LookupError:
                return request_id, session_id, mcp_transport

    if ctx is not None:
        request_id = ctx.request_id
        if hasattr(ctx, "request") and ctx.request is not None:
            request = ctx.request
            # Detect transport type by checking request characteristics
            if hasattr(request, "query_params") and request.query_params.get(
                "session_id"
            ):
                # SSE transport uses query parameter
                mcp_transport = "sse"
                session_id = request.query_params.get("session_id")
            elif hasattr(request, "headers") and request.headers.get("mcp-session-id"):
                # StreamableHTTP transport uses header
                mcp_transport = "http"
                session_id = request.headers.get("mcp-session-id")

    return request_id, session_id, mcp_transport


def _get_span_config(
    handler_type: str, item_name: str
) -> "tuple[str, str, str, Optional[str]]":
    """
    Get span configuration based on handler type.

    Returns:
        Tuple of (span_data_key, span_name, mcp_method_name, result_data_key)
        Note: result_data_key is None for resources
    """
    if handler_type == "tool":
        span_data_key = SPANDATA.MCP_TOOL_NAME
        mcp_method_name = "tools/call"
        result_data_key = SPANDATA.MCP_TOOL_RESULT_CONTENT
    elif handler_type == "prompt":
        span_data_key = SPANDATA.MCP_PROMPT_NAME
        mcp_method_name = "prompts/get"
        result_data_key = SPANDATA.MCP_PROMPT_RESULT_MESSAGE_CONTENT
    else:  # resource
        span_data_key = SPANDATA.MCP_RESOURCE_URI
        mcp_method_name = "resources/read"
        result_data_key = None  # Resources don't capture result content

    span_name = f"{mcp_method_name} {item_name}"
    return span_data_key, span_name, mcp_method_name, result_data_key


def _set_span_input_data(
    span: "Union[StreamedSpan, Span]",
    handler_name: str,
    span_data_key: str,
    mcp_method_name: str,
    arguments: "dict[str, Any]",
    request_id: "Optional[str]",
    session_id: "Optional[str]",
    mcp_transport: str,
) -> None:
    """Set input span data for MCP handlers."""

    # Set handler identifier
    _set_span_data_attribute(span, span_data_key, handler_name)
    _set_span_data_attribute(span, SPANDATA.MCP_METHOD_NAME, mcp_method_name)

    # Set transport/MCP transport type
    _set_span_data_attribute(
        span,
        SPANDATA.NETWORK_TRANSPORT,
        "pipe" if mcp_transport == "stdio" else "tcp",
    )
    _set_span_data_attribute(span, SPANDATA.MCP_TRANSPORT, mcp_transport)

    # Set request_id if provided
    if request_id:
        _set_span_data_attribute(span, SPANDATA.MCP_REQUEST_ID, request_id)

    # Set session_id if provided
    if session_id:
        _set_span_data_attribute(span, SPANDATA.MCP_SESSION_ID, session_id)

    # Set request arguments (excluding common request context objects)
    for k, v in arguments.items():
        _set_span_data_attribute(span, f"mcp.request.argument.{k}", safe_serialize(v))


def _extract_tool_result_content(result: "Any") -> "Any":
    """
    Extract meaningful content from MCP tool result.

    Tool handlers can return:
    - CallToolResult (mcp v2+): Has .content list and optional .structured_content
    - tuple (UnstructuredContent, StructuredContent): Return the structured content (dict)
    - dict (StructuredContent): Return as-is
    - list/Iterable (UnstructuredContent): Extract text from content blocks
    """
    if result is None:
        return None

    # Handle v2 CallToolResult-like objects (has .content list attribute)
    if hasattr(result, "content") and isinstance(
        getattr(result, "content", None), list
    ):
        # This is only present when a tool declares an output_schema
        structured = getattr(result, "structured_content", None)
        if structured is not None:
            return structured
        return _extract_text_from_content_blocks(result.content)

    # Handle CombinationContent: tuple of (UnstructuredContent, StructuredContent)
    if isinstance(result, tuple) and len(result) == 2:
        # Return the structured content (2nd element)
        return result[1]

    # Handle StructuredContent: dict
    if isinstance(result, dict):
        return result

    # Handle UnstructuredContent: iterable of ContentBlock objects
    if hasattr(result, "__iter__") and not isinstance(result, (str, bytes, dict)):
        return _extract_text_from_content_blocks(result)

    return result


def _extract_text_from_content_blocks(content_blocks: "Any") -> "Any":
    texts = []
    try:
        for item in content_blocks:
            if hasattr(item, "text"):
                texts.append(item.text)
            elif isinstance(item, dict) and "text" in item:
                texts.append(item["text"])
    except Exception:
        return content_blocks
    return " ".join(texts) if texts else content_blocks


def _set_span_output_data(
    span: "Union[StreamedSpan, Span]",
    result: "Any",
    result_data_key: "Optional[str]",
    handler_type: str,
) -> None:
    """Set output span data for MCP handlers."""
    if result is None:
        return

    # Get integration to check PII settings
    integration = sentry_sdk.get_client().get_integration(MCPIntegration)
    if integration is None:
        return

    # Check if we should include sensitive data
    should_include_data = should_send_default_pii() and integration.include_prompts

    # For tools, extract the meaningful content
    if handler_type == "tool":
        extracted = _extract_tool_result_content(result)
        if (
            extracted is not None
            and should_include_data
            and result_data_key is not None
        ):
            _set_span_data_attribute(span, result_data_key, safe_serialize(extracted))
            # Set content count if result is a dict
            if isinstance(extracted, dict):
                _set_span_data_attribute(
                    span, SPANDATA.MCP_TOOL_RESULT_CONTENT_COUNT, len(extracted)
                )
    elif handler_type == "prompt":
        # For prompts, count messages and set role/content only for single-message prompts
        try:
            messages: "Optional[list[str]]" = None
            message_count = 0

            # Check if result has messages attribute (GetPromptResult)
            if hasattr(result, "messages") and result.messages:
                messages = result.messages
                message_count = len(messages)
            # Also check if result is a dict with messages
            elif isinstance(result, dict) and result.get("messages"):
                messages = result["messages"]
                message_count = len(messages)

            # Always set message count if we found messages
            if message_count > 0:
                _set_span_data_attribute(
                    span, SPANDATA.MCP_PROMPT_RESULT_MESSAGE_COUNT, message_count
                )

            # Only set role and content for single-message prompts if PII is allowed
            if message_count == 1 and should_include_data and messages:
                first_message = messages[0]
                # Extract role
                role = None
                if hasattr(first_message, "role"):
                    role = first_message.role
                elif isinstance(first_message, dict) and "role" in first_message:
                    role = first_message["role"]

                if role:
                    _set_span_data_attribute(
                        span, SPANDATA.MCP_PROMPT_RESULT_MESSAGE_ROLE, role
                    )

                # Extract content text
                content_text = None
                if hasattr(first_message, "content"):
                    msg_content = first_message.content
                    # Content can be a TextContent object or similar
                    if hasattr(msg_content, "text"):
                        content_text = msg_content.text
                    elif isinstance(msg_content, dict) and "text" in msg_content:
                        content_text = msg_content["text"]
                    elif isinstance(msg_content, str):
                        content_text = msg_content
                elif isinstance(first_message, dict) and "content" in first_message:
                    msg_content = first_message["content"]
                    if isinstance(msg_content, dict) and "text" in msg_content:
                        content_text = msg_content["text"]
                    elif isinstance(msg_content, str):
                        content_text = msg_content

                if content_text and result_data_key is not None:
                    _set_span_data_attribute(span, result_data_key, content_text)
        except Exception:
            # Silently ignore if we can't extract message info
            pass
    # Resources don't capture result content (result_data_key is None)


# Handler data preparation and wrapping


def _is_v2_context(original_args: "tuple[Any, ...]") -> bool:
    """Check if original_args contains a v2 ServerRequestContext as the first element."""
    return (
        ServerRequestContext is not None
        and bool(original_args)
        and isinstance(original_args[0], ServerRequestContext)
    )


def _extract_handler_data_from_params(
    handler_type: str,
    params: "Any",
) -> "tuple[str, dict[str, Any]]":
    """
    Extract handler name and arguments from a v2 typed params object.

    In MCP SDK v2, handlers receive (ctx, params) where params is a typed
    Pydantic model (CallToolRequestParams, GetPromptRequestParams, etc.).
    """
    if handler_type == "tool":
        handler_name = getattr(params, "name", "unknown")
        arguments = getattr(params, "arguments", None) or {}
    elif handler_type == "prompt":
        handler_name = getattr(params, "name", "unknown")
        arguments = getattr(params, "arguments", None) or {}
        arguments = {"name": handler_name, **arguments}
    else:  # resource
        handler_name = str(getattr(params, "uri", "unknown"))
        arguments = {}

    return handler_name, arguments


def _extract_handler_data_from_args(
    handler_type: str,
    original_args: "tuple[Any, ...]",
    original_kwargs: "Optional[dict[str, Any]]" = None,
) -> "tuple[str, dict[str, Any]]":
    """
    Extract handler name and arguments from v1 positional args.

    In MCP SDK v1, handlers receive positional args:
    - Tool: (tool_name, arguments)
    - Prompt: (name, arguments)
    - Resource: (uri,)
    """
    original_kwargs = original_kwargs or {}

    if handler_type == "tool":
        if original_args:
            handler_name = original_args[0]
        elif original_kwargs.get("name"):
            handler_name = original_kwargs["name"]

        arguments = {}
        if len(original_args) > 1:
            arguments = original_args[1]
        elif original_kwargs.get("arguments"):
            arguments = original_kwargs["arguments"]

    elif handler_type == "prompt":
        if original_args:
            handler_name = original_args[0]
        elif original_kwargs.get("name"):
            handler_name = original_kwargs["name"]

        arguments = {}
        if len(original_args) > 1:
            arguments = original_args[1]
        elif original_kwargs.get("arguments"):
            arguments = original_kwargs["arguments"]

        arguments = {"name": handler_name, **(arguments or {})}

    else:  # resource
        handler_name = "unknown"
        if original_args:
            handler_name = str(original_args[0])
        elif original_kwargs.get("uri"):
            handler_name = str(original_kwargs["uri"])

        arguments = {}

    return handler_name, arguments


def _prepare_handler_data(
    handler_type: str,
    original_args: "tuple[Any, ...]",
    original_kwargs: "Optional[dict[str, Any]]" = None,
    params: "Optional[Any]" = None,
) -> "tuple[str, dict[str, Any], str, str, str, Optional[str]]":
    """
    Prepare common handler data for both v1 and v2 MCP SDK.

    Args:
        handler_type: "tool", "prompt", or "resource"
        original_args: Original positional args (v1 path)
        original_kwargs: Original keyword args (v1 path)
        params: Typed params object from v2 ServerRequestContext path

    Returns:
        Tuple of (handler_name, arguments, span_data_key, span_name, mcp_method_name, result_data_key)
    """
    if params is not None:
        handler_name, arguments = _extract_handler_data_from_params(
            handler_type, params
        )
    elif _is_v2_context(original_args):
        handler_name = "unknown"
        arguments = {}
    else:
        handler_name, arguments = _extract_handler_data_from_args(
            handler_type, original_args, original_kwargs
        )

    span_data_key, span_name, mcp_method_name, result_data_key = _get_span_config(
        handler_type, handler_name
    )

    return (
        handler_name,
        arguments,
        span_data_key,
        span_name,
        mcp_method_name,
        result_data_key,
    )


async def _handler_wrapper(
    handler_type: str,
    func: "Callable[..., Any]",
    original_args: "tuple[Any, ...]",
    original_kwargs: "Optional[dict[str, Any]]" = None,
    self: "Optional[Any]" = None,
    force_await: bool = True,
) -> "Any":
    """
    Wrapper for MCP handlers.

    Args:
        handler_type: "tool", "prompt", or "resource"
        func: The handler function to wrap
        original_args: Original arguments passed to the handler
        original_kwargs: Original keyword arguments passed to the handler
        self: Optional instance for bound methods
    """
    if original_kwargs is None:
        original_kwargs = {}

    # Detect v1 vs v2: MCP SDK v2 passes (ServerRequestContext, params) to handlers
    ctx: "Optional[Any]" = None
    params: "Optional[Any]" = None
    if (
        ServerRequestContext is not None
        and original_args
        and isinstance(original_args[0], ServerRequestContext)
    ):
        ctx = original_args[0]
        params = original_args[1] if len(original_args) > 1 else None

    (
        handler_name,
        arguments,
        span_data_key,
        span_name,
        mcp_method_name,
        result_data_key,
    ) = _prepare_handler_data(
        handler_type, original_args, original_kwargs, params=params
    )

    scopes = _get_active_http_scopes(ctx=ctx)

    isolation_scope_context: "ContextManager[Any]"
    current_scope_context: "ContextManager[Any]"

    if scopes is None:
        isolation_scope_context = nullcontext()
        current_scope_context = nullcontext()
    else:
        isolation_scope, current_scope = scopes

        isolation_scope_context = (
            nullcontext()
            if isolation_scope is None
            else sentry_sdk.scope.use_isolation_scope(isolation_scope)
        )
        current_scope_context = (
            nullcontext()
            if current_scope is None
            else sentry_sdk.scope.use_scope(current_scope)
        )

    # Get request ID, session ID, and transport from context
    request_id, session_id, mcp_transport = _get_request_context_data(ctx=ctx)

    span_streaming = has_span_streaming_enabled(sentry_sdk.get_client().options)

    # Start span and execute
    with isolation_scope_context:
        with current_scope_context:
            span_mgr: "Union[Span, StreamedSpan, ContextManager[None]]"
            if span_streaming:
                if sentry_sdk.traces.get_current_span() is not None:
                    span_mgr = sentry_sdk.traces.start_span(
                        name=span_name,
                        attributes={
                            "sentry.op": OP.MCP_SERVER,
                            "sentry.origin": MCPIntegration.origin,
                        },
                    )
                else:
                    span_mgr = nullcontext()
            else:
                span_mgr = get_start_span_function()(
                    op=OP.MCP_SERVER,
                    name=span_name,
                    origin=MCPIntegration.origin,
                )

            with span_mgr as span:
                if span is not None:
                    # Set input span data
                    _set_span_input_data(
                        span,
                        handler_name,
                        span_data_key,
                        mcp_method_name,
                        arguments,
                        request_id,
                        session_id,
                        mcp_transport,
                    )

                    # For resources, extract and set protocol
                    if handler_type == "resource":
                        uri = None
                        if params is not None:
                            uri = getattr(params, "uri", None)

                        # v1 scenario
                        if ServerRequestContext is None:
                            if original_args:
                                uri = original_args[0]
                            else:
                                uri = original_kwargs.get("uri")

                        protocol = None
                        if uri is not None and hasattr(uri, "scheme"):
                            protocol = uri.scheme
                        elif handler_name and "://" in handler_name:
                            protocol = handler_name.split("://")[0]
                        if protocol:
                            _set_span_data_attribute(
                                span, SPANDATA.MCP_RESOURCE_PROTOCOL, protocol
                            )

                try:
                    # Execute the async handler
                    if self is not None:
                        original_args = (self, *original_args)

                    result = func(*original_args, **original_kwargs)
                    if force_await or inspect.isawaitable(result):
                        result = await result

                except Exception as e:
                    # Set error flag for tools
                    if span is not None and handler_type == "tool":
                        _set_span_data_attribute(
                            span, SPANDATA.MCP_TOOL_RESULT_IS_ERROR, True
                        )
                    sentry_sdk.capture_exception(e)
                    raise

                if span is not None:
                    _set_span_output_data(span, result, result_data_key, handler_type)

    return result


def _create_instrumented_decorator(
    original_decorator: "Callable[..., Any]",
    handler_type: str,
    *decorator_args: "Any",
    **decorator_kwargs: "Any",
) -> "Callable[..., Any]":
    """
    Create an instrumented version of an MCP decorator.

    This function intercepts MCP decorators (like @server.call_tool()) and injects
    Sentry instrumentation into the handler registration flow. The returned decorator
    will:
    1. Receive the user's handler function
    2. Pass the instrumented version to the original MCP decorator

    This ensures that when the handler is called at runtime, it's already wrapped
    with Sentry spans and metrics collection.

    Args:
        original_decorator: The original MCP decorator method (e.g., Server.call_tool)
        handler_type: "tool", "prompt", or "resource" - determines span configuration
        decorator_args: Positional arguments to pass to the original decorator (e.g., self)
        decorator_kwargs: Keyword arguments to pass to the original decorator

    Returns:
        A decorator function that instruments handlers before registering them
    """

    def instrumented_decorator(func: "Callable[..., Any]") -> "Callable[..., Any]":
        @wraps(func)
        async def wrapper(*args: "Any") -> "Any":
            return await _handler_wrapper(handler_type, func, args, force_await=False)

        # Then register it with the original MCP decorator
        return original_decorator(*decorator_args, **decorator_kwargs)(wrapper)

    return instrumented_decorator


_METHOD_TO_HANDLER_TYPE = {
    "tools/call": "tool",
    "prompts/get": "prompt",
    "resources/read": "resource",
}

# In MCP SDK v2, tool/prompt/resource handlers are most commonly registered via
# the Server(...) constructor kwargs rather than add_request_handler. The in-tree
# high-level MCPServer also wires its handlers through these kwargs.
_KWARG_TO_HANDLER_TYPE = {
    "on_call_tool": "tool",
    "on_get_prompt": "prompt",
    "on_read_resource": "resource",
}


def _patch_lowlevel_server() -> None:
    """
    Patches the mcp.server.lowlevel.Server class to instrument handler execution.
    """
    if MCP_PACKAGE_VERSION and MCP_PACKAGE_VERSION >= (2, 0, 0):
        _patch_lowlevel_server_v2()
    else:
        _patch_lowlevel_server_v1()


def _patch_lowlevel_server_v1() -> None:
    """Patches v1 Server decorator methods (call_tool, get_prompt, read_resource)."""
    # Patch call_tool decorator
    original_call_tool = Server.call_tool  # type: ignore[attr-defined]

    def patched_call_tool(
        self: "Server", **kwargs: "Any"
    ) -> "Callable[[Callable[..., Any]], Callable[..., Any]]":
        """Patched version of Server.call_tool that adds Sentry instrumentation."""
        return lambda func: _create_instrumented_decorator(
            original_call_tool, "tool", self, **kwargs
        )(func)

    Server.call_tool = patched_call_tool  # type: ignore[attr-defined]

    # Patch get_prompt decorator
    original_get_prompt = Server.get_prompt  # type: ignore[attr-defined]

    def patched_get_prompt(
        self: "Server",
    ) -> "Callable[[Callable[..., Any]], Callable[..., Any]]":
        """Patched version of Server.get_prompt that adds Sentry instrumentation."""
        return lambda func: _create_instrumented_decorator(
            original_get_prompt, "prompt", self
        )(func)

    Server.get_prompt = patched_get_prompt  # type: ignore[attr-defined]

    # Patch read_resource decorator
    original_read_resource = Server.read_resource  # type: ignore[attr-defined]

    def patched_read_resource(
        self: "Server",
    ) -> "Callable[[Callable[..., Any]], Callable[..., Any]]":
        """Patched version of Server.read_resource that adds Sentry instrumentation."""
        return lambda func: _create_instrumented_decorator(
            original_read_resource, "resource", self
        )(func)

    Server.read_resource = patched_read_resource  # type: ignore[attr-defined]


def _wrap_v2_handler(
    handler_type: str, handler: "Callable[..., Any]"
) -> "Callable[..., Any]":
    """Wrap a v2 (ctx, params) handler with Sentry instrumentation.

    Idempotent: an already-wrapped handler is returned unchanged so handlers
    registered through more than one path (e.g. MCPServer building a Server)
    are not double-wrapped.
    """
    if getattr(handler, "__sentry_mcp_wrapped__", False):
        return handler

    @wraps(handler)
    async def wrapper(*args: "Any", **kwargs: "Any") -> "Any":
        return await _handler_wrapper(
            handler_type, handler, args, kwargs, force_await=False
        )

    wrapper.__sentry_mcp_wrapped__ = True  # type: ignore[attr-defined]
    return wrapper


def _patch_lowlevel_server_v2() -> None:
    """Patches the v2 Server to wrap tool/prompt/resource handlers.

    Handlers can be registered either via the Server(...) constructor kwargs
    (on_call_tool/on_get_prompt/on_read_resource) — the path the in-tree
    MCPServer and most lowlevel examples use — or via add_request_handler.
    Both are patched.
    """
    original_init = Server.__init__

    @wraps(original_init)
    def patched_init(self: "Server", *args: "Any", **kwargs: "Any") -> None:
        for kwarg, handler_type in _KWARG_TO_HANDLER_TYPE.items():
            handler = kwargs.get(kwarg)
            if handler is not None:
                kwargs[kwarg] = _wrap_v2_handler(handler_type, handler)
        original_init(self, *args, **kwargs)

    Server.__init__ = patched_init  # type: ignore[method-assign]

    original_add_request_handler = Server.add_request_handler

    def patched_add_request_handler(
        self: "Server",
        method: str,
        params_type: "Any",
        handler: "Callable[..., Any]",
        *args: "Any",
        **kwargs: "Any",
    ) -> None:
        handler_type = _METHOD_TO_HANDLER_TYPE.get(method)
        if handler_type is not None:
            handler = _wrap_v2_handler(handler_type, handler)

        original_add_request_handler(
            self, method, params_type, handler, *args, **kwargs
        )

    Server.add_request_handler = patched_add_request_handler  # type: ignore[method-assign]


def _patch_handle_request() -> None:
    original_handle_request = StreamableHTTPServerTransport.handle_request

    @wraps(original_handle_request)
    async def patched_handle_request(
        self: "StreamableHTTPServerTransport",
        scope: "Scope",
        receive: "Receive",
        send: "Send",
    ) -> None:
        scope.setdefault("state", {})["sentry_sdk.isolation_scope"] = (
            sentry_sdk.get_isolation_scope()
        )
        scope["state"]["sentry_sdk.current_scope"] = sentry_sdk.get_current_scope()
        await original_handle_request(self, scope, receive, send)

    StreamableHTTPServerTransport.handle_request = patched_handle_request  # type: ignore[method-assign]


def _patch_fastmcp() -> None:
    """
    Patches the standalone fastmcp package's FastMCP class.

    The standalone fastmcp package (v2.14.0+) registers its own handlers for
    prompts and resources directly, bypassing the Server decorators we patch.
    This function patches the _get_prompt_mcp and _read_resource_mcp methods
    to add instrumentation for those handlers.
    """
    if FastMCP is not None and hasattr(FastMCP, "_get_prompt_mcp"):
        original_get_prompt_mcp = FastMCP._get_prompt_mcp

        @wraps(original_get_prompt_mcp)
        async def patched_get_prompt_mcp(
            self: "Any", *args: "Any", **kwargs: "Any"
        ) -> "Any":
            return await _handler_wrapper(
                "prompt",
                original_get_prompt_mcp,
                args,
                kwargs,
                self,
            )

        FastMCP._get_prompt_mcp = patched_get_prompt_mcp

    if FastMCP is not None and hasattr(FastMCP, "_read_resource_mcp"):
        original_read_resource_mcp = FastMCP._read_resource_mcp

        @wraps(original_read_resource_mcp)
        async def patched_read_resource_mcp(
            self: "Any", *args: "Any", **kwargs: "Any"
        ) -> "Any":
            return await _handler_wrapper(
                "resource",
                original_read_resource_mcp,
                args,
                kwargs,
                self,
            )

        FastMCP._read_resource_mcp = patched_read_resource_mcp



import gc
import os
import uuid
from unittest import mock
from unittest.mock import MagicMock

import pytest

import sentry_sdk
from sentry_sdk import set_measurement, start_span, start_transaction
from sentry_sdk.consts import MATCH_ALL
from sentry_sdk.traces import StreamedSpan
from sentry_sdk.tracing import Span, Transaction
from sentry_sdk.tracing_utils import should_propagate_trace
from sentry_sdk.utils import Dsn
from tests.conftest import ApproxDict


def test_span_trimming(sentry_init, capture_events):
    sentry_init(traces_sample_rate=1.0, _experiments={"max_spans": 3})
    events = capture_events()

    with start_transaction(name="hi"):
        for i in range(10):
            with start_span(op="foo{}".format(i)):
                pass

    (event,) = events

    assert len(event["spans"]) == 3

    span1, span2, span3 = event["spans"]
    assert span1["op"] == "foo0"
    assert span2["op"] == "foo1"
    assert span3["op"] == "foo2"

    assert event["_meta"]["spans"][""]["len"] == 10
    assert "_dropped_spans" not in event
    assert "dropped_spans" not in event


def test_span_trimming_produces_client_report(
    sentry_init, capture_events, capture_record_lost_event_calls
):
    sentry_init(traces_sample_rate=1.0, _experiments={"max_spans": 3})
    events = capture_events()
    record_lost_event_calls = capture_record_lost_event_calls()

    with start_transaction(name="hi"):
        for i in range(10):
            with start_span(op="foo{}".format(i)):
                pass

    (event,) = events

    assert len(event["spans"]) == 3

    # 7 spans were dropped (10 total - 3 kept = 7 dropped)
    assert ("buffer_overflow", "span", None, 7) in record_lost_event_calls


def test_span_data_scrubbing_and_trimming(sentry_init, capture_events):
    sentry_init(traces_sample_rate=1.0, _experiments={"max_spans": 3})
    events = capture_events()

    with start_transaction(name="hi"):
        with start_span(op="foo", name="bar") as span:
            span.set_data("password", "secret")
            span.set_data("datafoo", "databar")

        for i in range(10):
            with start_span(op="foo{}".format(i)):
                pass

    (event,) = events
    assert event["spans"][0]["data"] == ApproxDict(
        {"password": "[Filtered]", "datafoo": "databar"}
    )
    assert event["_meta"]["spans"] == {
        "0": {"data": {"password": {"": {"rem": [["!config", "s"]]}}}},
        "": {"len": 11},
    }


def test_transaction_naming(sentry_init, capture_events):
    sentry_init(traces_sample_rate=1.0)
    events = capture_events()

    # default name in event if no name is passed
    with start_transaction() as transaction:
        pass
    assert len(events) == 1
    assert events[0]["transaction"] == "<unlabeled transaction>"

    # the name can be set once the transaction's already started
    with start_transaction() as transaction:
        transaction.name = "name-known-after-transaction-started"
    assert len(events) == 2
    assert events[1]["transaction"] == "name-known-after-transaction-started"

    # passing in a name works, too
    with start_transaction(name="a"):
        pass
    assert len(events) == 3
    assert events[2]["transaction"] == "a"


def test_transaction_data(sentry_init, capture_events):
    sentry_init(traces_sample_rate=1.0)
    events = capture_events()

    with start_transaction(name="test-transaction"):
        span_or_tx = sentry_sdk.get_current_span()
        span_or_tx.set_data("foo", "bar")
        with start_span(op="test-span") as span:
            span.set_data("spanfoo", "spanbar")

    assert len(events) == 1

    transaction = events[0]
    transaction_data = transaction["contexts"]["trace"]["data"]

    assert "data" not in transaction.keys()
    assert transaction_data.items() >= {"foo": "bar"}.items()

    assert len(transaction["spans"]) == 1

    span = transaction["spans"][0]
    span_data = span["data"]

    assert "contexts" not in span.keys()
    assert span_data.items() >= {"spanfoo": "spanbar"}.items()


def test_start_transaction(sentry_init):
    sentry_init(traces_sample_rate=1.0)

    # you can have it start a transaction for you
    result1 = start_transaction(
        name="/interactions/other-dogs/new-dog", op="greeting.sniff"
    )
    assert isinstance(result1, Transaction)
    assert result1.name == "/interactions/other-dogs/new-dog"
    assert result1.op == "greeting.sniff"

    # or you can pass it an already-created transaction
    preexisting_transaction = Transaction(
        name="/interactions/other-dogs/new-dog", op="greeting.sniff"
    )
    result2 = start_transaction(preexisting_transaction)
    assert result2 is preexisting_transaction


def test_finds_transaction_on_scope(sentry_init):
    sentry_init(traces_sample_rate=1.0)

    transaction = start_transaction(name="dogpark")

    scope = sentry_sdk.get_current_scope()

    # See note in Scope class re: getters and setters of the `transaction`
    # property. For the moment, assigning to scope.transaction merely sets the
    # transaction name, rather than putting the transaction on the scope, so we
    # have to assign to _span directly.
    scope._span = transaction

    # Reading scope.property, however, does what you'd expect, and returns the
    # transaction on the scope.
    assert scope.transaction is not None
    assert isinstance(scope.transaction, Transaction)
    assert scope.transaction.name == "dogpark"

    # If the transaction is also set as the span on the scope, it can be found
    # by accessing _span, too.
    assert scope._span is not None
    assert isinstance(scope._span, Transaction)
    assert scope._span.name == "dogpark"


def test_finds_segment_on_scope(sentry_init):
    sentry_init(
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )

    with sentry_sdk.traces.start_span(name="dogpark"):
        scope = sentry_sdk.get_current_scope()
        assert scope.streamed_span is not None
        assert isinstance(scope.streamed_span, StreamedSpan)
        assert scope.streamed_span.name == "dogpark"

        assert scope._span is not None
        assert isinstance(scope._span, StreamedSpan)
        assert scope._span.name == "dogpark"


def test_finds_transaction_when_descendent_span_is_on_scope(
    sentry_init,
):
    sentry_init(traces_sample_rate=1.0)

    transaction = start_transaction(name="dogpark")
    child_span = transaction.start_child(op="sniffing")

    scope = sentry_sdk.get_current_scope()
    scope._span = child_span

    # this is the same whether it's the transaction itself or one of its
    # decedents directly attached to the scope
    assert scope.transaction is not None
    assert isinstance(scope.transaction, Transaction)
    assert scope.transaction.name == "dogpark"

    # here we see that it is in fact the span on the scope, rather than the
    # transaction itself
    assert scope._span is not None
    assert isinstance(scope._span, Span)
    assert scope._span.op == "sniffing"


def test_finds_orphan_span_on_scope(sentry_init):
    # this is deprecated behavior which may be removed at some point (along with
    # the start_span function)
    sentry_init(traces_sample_rate=1.0)

    span = start_span(op="sniffing")

    scope = sentry_sdk.get_current_scope()
    scope._span = span

    assert scope._span is not None
    assert isinstance(scope._span, Span)
    assert scope._span.op == "sniffing"


def test_finds_non_orphan_span_on_scope(sentry_init):
    sentry_init(traces_sample_rate=1.0)

    transaction = start_transaction(name="dogpark")
    child_span = transaction.start_child(op="sniffing")

    scope = sentry_sdk.get_current_scope()
    scope._span = child_span

    assert scope._span is not None
    assert isinstance(scope._span, Span)
    assert scope._span.op == "sniffing"


def test_finds_non_orphan_span_on_scope_span_streaming(sentry_init):
    sentry_init(
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )

    segment = sentry_sdk.traces.start_span(name="dogpark")
    sentry_sdk.traces.start_span(name="sniffing", parent_span=segment)

    scope = sentry_sdk.get_current_scope()

    assert scope._span is not None
    assert isinstance(scope._span, StreamedSpan)
    assert scope._span.name == "sniffing"


def test_circular_references(monkeypatch, sentry_init, request):
    # TODO: We discovered while writing this test about transaction/span
    # reference cycles that there's actually also a circular reference in
    # `serializer.py`, between the functions `_serialize_node` and
    # `_serialize_node_impl`, both of which are defined inside of the main
    # `serialize` function, and each of which calls the other one. For now, in
    # order to avoid having those ref cycles give us a false positive here, we
    # can mock out `serialize`. In the long run, though, we should probably fix
    # that. (Whenever we do work on fixing it, it may be useful to add
    #
    #     gc.set_debug(gc.DEBUG_LEAK)
    #     request.addfinalizer(lambda: gc.set_debug(~gc.DEBUG_LEAK))
    #
    # immediately after the initial collection below, so we can see what new
    # objects the garbage collector has to clean up once `transaction.finish` is
    # called and the serializer runs.)
    monkeypatch.setattr(
        sentry_sdk.client,
        "serialize",
        mock.Mock(
            return_value=None,
        ),
    )

    # In certain versions of python, in some environments (specifically, python
    # 3.4 when run in GH Actions), we run into a `ctypes` bug which creates
    # circular references when `uuid4()` is called, as happens when we're
    # generating event ids. Mocking it with an implementation which doesn't use
    # the `ctypes` function lets us avoid having false positives when garbage
    # collecting. See https://bugs.python.org/issue20519.
    monkeypatch.setattr(
        uuid,
        "uuid4",
        mock.Mock(
            return_value=uuid.UUID(bytes=os.urandom(16)),
        ),
    )

    gc.disable()
    request.addfinalizer(gc.enable)

    sentry_init(traces_sample_rate=1.0)

    # Make sure that we're starting with a clean slate before we start creating
    # transaction/span reference cycles
    gc.collect()

    dogpark_transaction = start_transaction(name="dogpark")
    sniffing_span = dogpark_transaction.start_child(op="sniffing")
    wagging_span = dogpark_transaction.start_child(op="wagging")

    # At some point, you have to stop sniffing - there are balls to chase! - so finish
    # this span while the dogpark transaction is still open
    sniffing_span.finish()

    # The wagging, however, continues long past the dogpark, so that span will
    # NOT finish before the transaction ends. (Doing it in this order proves
    # that both finished and unfinished spans get their cycles broken.)
    dogpark_transaction.finish()

    # Eventually you gotta sleep...
    wagging_span.finish()

    # assuming there are no cycles by this point, these should all be able to go
    # out of scope and get their memory deallocated without the garbage
    # collector having anything to do
    del sniffing_span
    del wagging_span
    del dogpark_transaction

    assert gc.collect() == 0


def test_circular_references_span_streaming(monkeypatch, sentry_init, request):
    gc.disable()
    request.addfinalizer(gc.enable)

    sentry_init(
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )

    # Make sure that we're starting with a clean slate before we start creating
    # transaction/span reference cycles
    gc.collect()

    dogpark_segment = sentry_sdk.traces.start_span(name="dogpark")
    sniffing_span = sentry_sdk.traces.start_span(
        name="sniffing", parent_span=dogpark_segment
    )
    wagging_span = sentry_sdk.traces.start_span(
        name="wagging", parent_span=dogpark_segment
    )

    # At some point, you have to stop sniffing - there are balls to chase! - so finish
    # this span while the dogpark transaction is still open
    sniffing_span.end()

    # The wagging, however, continues long past the dogpark, so that span will
    # NOT finish before the transaction ends. (Doing it in this order proves
    # that both finished and unfinished spans get their cycles broken.)
    dogpark_segment.end()

    # Eventually you gotta sleep...
    wagging_span.end()

    # assuming there are no cycles by this point, these should all be able to go
    # out of scope and get their memory deallocated without the garbage
    # collector having anything to do
    del sniffing_span
    del wagging_span
    del dogpark_segment

    assert gc.collect() == 0


def test_set_measurement(sentry_init, capture_events):
    sentry_init(traces_sample_rate=1.0)

    events = capture_events()

    transaction = start_transaction(name="measuring stuff")

    with pytest.raises(TypeError):
        transaction.set_measurement()

    with pytest.raises(TypeError):
        transaction.set_measurement("metric.foo")

    transaction.set_measurement("metric.foo", 123)
    transaction.set_measurement("metric.bar", 456, unit="second")
    transaction.set_measurement("metric.baz", 420.69, unit="custom")
    transaction.set_measurement("metric.foobar", 12, unit="percent")
    transaction.set_measurement("metric.foobar", 17.99, unit="percent")

    transaction.finish()

    (event,) = events
    assert event["measurements"]["metric.foo"] == {"value": 123, "unit": ""}
    assert event["measurements"]["metric.bar"] == {"value": 456, "unit": "second"}
    assert event["measurements"]["metric.baz"] == {"value": 420.69, "unit": "custom"}
    assert event["measurements"]["metric.foobar"] == {"value": 17.99, "unit": "percent"}


def test_set_measurement_public_api(sentry_init, capture_events):
    sentry_init(traces_sample_rate=1.0)

    events = capture_events()

    with start_transaction(name="measuring stuff"):
        set_measurement("metric.foo", 123)
        set_measurement("metric.bar", 456, unit="second")

    (event,) = events
    assert event["measurements"]["metric.foo"] == {"value": 123, "unit": ""}
    assert event["measurements"]["metric.bar"] == {"value": 456, "unit": "second"}


def test_set_measurement_deprecated(sentry_init):
    sentry_init(traces_sample_rate=1.0)

    with start_transaction(name="measuring stuff") as trx:
        with pytest.warns(DeprecationWarning):
            set_measurement("metric.foo", 123)

        with pytest.warns(DeprecationWarning):
            trx.set_measurement("metric.bar", 456)

        with start_span(op="measuring span") as span:
            with pytest.warns(DeprecationWarning):
                span.set_measurement("metric.baz", 420.69, unit="custom")


def test_set_meaurement_compared_to_set_data(sentry_init, capture_events):
    """
    This is just a test to see the difference
    between measurements and data in the resulting event payload.
    """
    sentry_init(traces_sample_rate=1.0)

    events = capture_events()

    with start_transaction(name="measuring stuff") as transaction:
        transaction.set_measurement("metric.foo", 123)
        transaction.set_data("metric.bar", 456)

        with start_span(op="measuring span") as span:
            span.set_measurement("metric.baz", 420.69, unit="custom")
            span.set_data("metric.qux", 789)

    (event,) = events
    assert event["measurements"]["metric.foo"] == {"value": 123, "unit": ""}
    assert event["contexts"]["trace"]["data"]["metric.bar"] == 456
    assert event["spans"][0]["measurements"]["metric.baz"] == {
        "value": 420.69,
        "unit": "custom",
    }
    assert event["spans"][0]["data"]["metric.qux"] == 789


@pytest.mark.parametrize(
    "trace_propagation_targets,url,expected_propagation_decision",
    [
        (None, "http://example.com", False),
        ([], "http://example.com", False),
        ([MATCH_ALL], "http://example.com", True),
        (["localhost"], "localhost:8443/api/users", True),
        (["localhost"], "http://localhost:8443/api/users", True),
        (["localhost"], "mylocalhost:8080/api/users", True),
        ([r"^/api"], "/api/envelopes", True),
        ([r"^/api"], "/backend/api/envelopes", False),
        ([r"myApi.com/v[2-4]"], "myApi.com/v2/projects", True),
        ([r"myApi.com/v[2-4]"], "myApi.com/v1/projects", False),
        ([r"https:\/\/.*"], "https://example.com", True),
        (
            [r"https://.*"],
            "https://example.com",
            True,
        ),  # to show escaping is not needed
        ([r"https://.*"], "http://example.com/insecure/", False),
    ],
)
def test_should_propagate_trace(
    trace_propagation_targets, url, expected_propagation_decision
):
    client = MagicMock()

    # This test assumes the urls are not Sentry URLs. Use test_should_propagate_trace_to_sentry for sentry URLs.
    client.is_sentry_url = lambda _: False

    client.options = {"trace_propagation_targets": trace_propagation_targets}
    client.transport = MagicMock()
    client.transport.parsed_dsn = Dsn("https://bla@xxx.sentry.io/12312012")

    assert should_propagate_trace(client, url) == expected_propagation_decision


@pytest.mark.parametrize(
    "dsn,url,expected_propagation_decision",
    [
        (
            "https://dogsarebadatkeepingsecrets@squirrelchasers.ingest.sentry.io/12312012",
            "http://example.com",
            True,
        ),
        (
            "https://dogsarebadatkeepingsecrets@squirrelchasers.ingest.sentry.io/12312012",
            "https://dogsarebadatkeepingsecrets@squirrelchasers.ingest.sentry.io/12312012",
            False,
        ),
        (
            "https://dogsarebadatkeepingsecrets@squirrelchasers.ingest.sentry.io/12312012",
            "http://squirrelchasers.ingest.sentry.io/12312012",
            False,
        ),
        (
            "https://dogsarebadatkeepingsecrets@squirrelchasers.ingest.sentry.io/12312012",
            "http://ingest.sentry.io/12312012",
            True,
        ),
        (
            "https://abc@localsentry.example.com/12312012",
            "http://localsentry.example.com",
            False,
        ),
    ],
)
def test_should_propagate_trace_to_sentry(
    sentry_init, dsn, url, expected_propagation_decision
):
    sentry_init(
        dsn=dsn,
        traces_sample_rate=1.0,
    )

    client = sentry_sdk.get_client()
    client.transport.parsed_dsn = Dsn(dsn)

    assert should_propagate_trace(client, url) == expected_propagation_decision


def test_start_transaction_updates_scope_name_source(sentry_init):
    sentry_init(traces_sample_rate=1.0)

    scope = sentry_sdk.get_current_scope()

    with start_transaction(name="foobar", source="route"):
        assert scope._transaction == "foobar"
        assert scope._transaction_info == {"source": "route"}


@pytest.mark.parametrize("sampled", (True, None))
def test_transaction_dropped_debug_not_started(sentry_init, sampled):
    sentry_init(traces_sample_rate=1.0)

    tx = Transaction(sampled=sampled)

    with mock.patch("sentry_sdk.tracing.logger") as mock_logger:
        with tx:
            pass

    mock_logger.debug.assert_any_call(
        "Discarding transaction because it was not started with sentry_sdk.start_transaction"
    )

    with pytest.raises(AssertionError):
        # We should NOT see the "sampled = False" message here
        mock_logger.debug.assert_any_call(
            "Discarding transaction because sampled = False"
        )


def test_transaction_dropped_sampled_false(sentry_init):
    sentry_init(traces_sample_rate=1.0)

    tx = Transaction(sampled=False)

    with mock.patch("sentry_sdk.tracing.logger") as mock_logger:
        with sentry_sdk.start_transaction(tx):
            pass

    mock_logger.debug.assert_any_call("Discarding transaction because sampled = False")

    with pytest.raises(AssertionError):
        # We should not see the "not started" message here
        mock_logger.debug.assert_any_call(
            "Discarding transaction because it was not started with sentry_sdk.start_transaction"
        )


def test_transaction_not_started_warning(sentry_init):
    sentry_init(traces_sample_rate=1.0)

    tx = Transaction()

    with mock.patch("sentry_sdk.tracing.logger") as mock_logger:
        with tx:
            pass

    mock_logger.debug.assert_any_call(
        "Transaction was entered without being started with sentry_sdk.start_transaction."
        "The transaction will not be sent to Sentry. To fix, start the transaction by"
        "passing it to sentry_sdk.start_transaction."
    )


def test_span_set_data_update_data(sentry_init, capture_events):
    sentry_init(traces_sample_rate=1.0)

    events = capture_events()

    with sentry_sdk.start_transaction(name="test-transaction"):
        with start_span(op="test-span") as span:
            span.set_data("key0", "value0")
            span.set_data("key1", "value1")

            span.update_data(
                {
                    "key1": "updated-value1",
                    "key2": "value2",
                    "key3": "value3",
                }
            )

    (event,) = events
    span = event["spans"][0]

    assert span["data"] == {
        "key0": "value0",
        "key1": "updated-value1",
        "key2": "value2",
        "key3": "value3",
        "thread.id": mock.ANY,
        "thread.name": mock.ANY,
    }


def test_update_current_span(sentry_init, capture_events):
    sentry_init(traces_sample_rate=1.0)

    events = capture_events()

    with sentry_sdk.start_transaction(name="test-transaction"):
        with start_span(op="test-span-op", name="test-span-name"):
            sentry_sdk.update_current_span(
                op="updated-span-op",
                name="updated-span-name",
                attributes={
                    "key0": "value0",
                    "key1": "value1",
                },
            )

            sentry_sdk.update_current_span(
                op="updated-span-op-2",
            )

            sentry_sdk.update_current_span(
                name="updated-span-name-3",
            )

            sentry_sdk.update_current_span(
                attributes={
                    "key1": "updated-value-4",
                    "key2": "value2",
                },
            )

    (event,) = events
    span = event["spans"][0]

    assert span["op"] == "updated-span-op-2"
    assert span["description"] == "updated-span-name-3"
    assert span["data"] == {
        "key0": "value0",
        "key1": "updated-value-4",
        "key2": "value2",
        "thread.id": mock.ANY,
        "thread.name": mock.ANY,
    }



from typing import TYPE_CHECKING

import sentry_sdk
from sentry_sdk.consts import OP, SPANDATA, SPANSTATUS
from sentry_sdk.scope import should_send_default_pii
from sentry_sdk.traces import SpanStatus, StreamedSpan
from sentry_sdk.tracing_utils import has_span_streaming_enabled

from ..consts import SPAN_ORIGIN
from ..utils import _set_agent_data

if TYPE_CHECKING:
    from typing import Any, Union

    import agents


def execute_tool_span(
    tool: "agents.Tool", *args: "Any", **kwargs: "Any"
) -> "Union[sentry_sdk.tracing.Span, StreamedSpan]":
    span_streaming = has_span_streaming_enabled(sentry_sdk.get_client().options)
    if span_streaming:
        span = sentry_sdk.traces.start_span(
            name=f"execute_tool {tool.name}",
            attributes={
                "sentry.op": OP.GEN_AI_EXECUTE_TOOL,
                "sentry.origin": SPAN_ORIGIN,
                SPANDATA.GEN_AI_OPERATION_NAME: "execute_tool",
                SPANDATA.GEN_AI_TOOL_NAME: tool.name,
                SPANDATA.GEN_AI_TOOL_DESCRIPTION: tool.description,
            },
        )

        set_on_span = span.set_attribute
    else:
        span = sentry_sdk.start_span(
            op=OP.GEN_AI_EXECUTE_TOOL,
            name=f"execute_tool {tool.name}",
            origin=SPAN_ORIGIN,
        )

        span.set_data(SPANDATA.GEN_AI_OPERATION_NAME, "execute_tool")

        span.set_data(SPANDATA.GEN_AI_TOOL_NAME, tool.name)
        span.set_data(SPANDATA.GEN_AI_TOOL_DESCRIPTION, tool.description)

        set_on_span = span.set_data

    if should_send_default_pii():
        input = args[1]
        set_on_span(SPANDATA.GEN_AI_TOOL_INPUT, input)

    return span


def update_execute_tool_span(
    span: "Union[sentry_sdk.tracing.Span, StreamedSpan]",
    agent: "agents.Agent",
    tool: "agents.Tool",
    result: "Any",
) -> None:
    _set_agent_data(span, agent)

    if isinstance(result, str) and result.startswith(
        "An error occurred while running the tool"
    ):
        if isinstance(span, StreamedSpan):
            span.status = SpanStatus.ERROR
        else:
            span.set_status(SPANSTATUS.INTERNAL_ERROR)

    set_on_span = (
        span.set_attribute if isinstance(span, StreamedSpan) else span.set_data
    )

    if should_send_default_pii():
        set_on_span(SPANDATA.GEN_AI_TOOL_OUTPUT, result)

    # Add conversation ID from agent
    conv_id = getattr(agent, "_sentry_conversation_id", None)
    if conv_id:
        set_on_span(SPANDATA.GEN_AI_CONVERSATION_ID, conv_id)



import pytest

import sentry_sdk


def test_standalone_span_iter_headers(sentry_init):
    sentry_init(traces_sample_rate=1.0)

    with sentry_sdk.start_span(op="test") as span:
        with pytest.raises(StopIteration):
            # We should not have any propagation headers
            next(span.iter_headers())


def test_span_in_span_iter_headers(sentry_init):
    sentry_init(traces_sample_rate=1.0)

    with sentry_sdk.start_span(op="test"):
        with sentry_sdk.start_span(op="test2") as span_inner:
            with pytest.raises(StopIteration):
                # We should not have any propagation headers
                next(span_inner.iter_headers())


def test_span_in_transaction(sentry_init):
    sentry_init(traces_sample_rate=1.0)

    with sentry_sdk.start_transaction(op="test"):
        with sentry_sdk.start_span(op="test2") as span:
            # Ensure the headers are there
            next(span.iter_headers())


def test_span_in_transaction_span_streaming(sentry_init):
    sentry_init(traces_sample_rate=1.0, trace_lifecycle="stream")

    with sentry_sdk.traces.start_span(name="test"):
        with sentry_sdk.traces.start_span(name="test2") as span:
            # Ensure the headers are there
            next(span._iter_headers())


def test_span_in_span_in_transaction(sentry_init):
    sentry_init(traces_sample_rate=1.0)

    with sentry_sdk.start_transaction(op="test"):
        with sentry_sdk.start_span(op="test2"):
            with sentry_sdk.start_span(op="test3") as span_inner:
                # Ensure the headers are there
                next(span_inner.iter_headers())


def test_span_in_span_in_transaction_span_streaming(sentry_init):
    sentry_init(traces_sample_rate=1.0, trace_lifecycle="stream")

    with sentry_sdk.traces.start_span(name="test"):
        with sentry_sdk.traces.start_span(name="test2"):
            with sentry_sdk.traces.start_span(name="test3") as span_inner:
                # Ensure the headers are there
                next(span_inner._iter_headers())



import datetime
from unittest import mock
from unittest.mock import MagicMock

import pytest
from celery.schedules import crontab, schedule

from sentry_sdk.crons import MonitorStatus
from sentry_sdk.integrations.celery.beat import (
    _get_headers,
    _get_monitor_config,
    _patch_beat_apply_entry,
    _patch_redbeat_apply_async,
    crons_task_failure,
    crons_task_retry,
    crons_task_success,
)
from sentry_sdk.integrations.celery.utils import _get_humanized_interval


def test_get_headers():
    fake_task = MagicMock()
    fake_task.request = {
        "bla": "blub",
        "foo": "bar",
    }

    assert _get_headers(fake_task) == {}

    fake_task.request.update(
        {
            "headers": {
                "bla": "blub",
            },
        }
    )

    assert _get_headers(fake_task) == {"bla": "blub"}

    fake_task.request.update(
        {
            "headers": {
                "headers": {
                    "tri": "blub",
                    "bar": "baz",
                },
                "bla": "blub",
            },
        }
    )

    assert _get_headers(fake_task) == {"bla": "blub", "tri": "blub", "bar": "baz"}


@pytest.mark.parametrize(
    "seconds, expected_tuple",
    [
        (0, (0, "second")),
        (1, (1, "second")),
        (0.00001, (0, "second")),
        (59, (59, "second")),
        (60, (1, "minute")),
        (100, (1, "minute")),
        (1000, (16, "minute")),
        (10000, (2, "hour")),
        (100000, (1, "day")),
        (100000000, (1157, "day")),
    ],
)
def test_get_humanized_interval(seconds, expected_tuple):
    assert _get_humanized_interval(seconds) == expected_tuple


def test_crons_task_success():
    fake_task = MagicMock()
    fake_task.request = {
        "headers": {
            "sentry-monitor-slug": "test123",
            "sentry-monitor-check-in-id": "1234567890",
            "sentry-monitor-start-timestamp-s": 200.1,
            "sentry-monitor-config": {
                "schedule": {
                    "type": "interval",
                    "value": 3,
                    "unit": "day",
                },
                "timezone": "Europe/Vienna",
            },
            "sentry-monitor-some-future-key": "some-future-value",
        },
    }

    with mock.patch(
        "sentry_sdk.integrations.celery.beat.capture_checkin"
    ) as mock_capture_checkin:
        with mock.patch(
            "sentry_sdk.integrations.celery.beat._now_seconds_since_epoch",
            return_value=500.5,
        ):
            crons_task_success(fake_task)

            mock_capture_checkin.assert_called_once_with(
                monitor_slug="test123",
                monitor_config={
                    "schedule": {
                        "type": "interval",
                        "value": 3,
                        "unit": "day",
                    },
                    "timezone": "Europe/Vienna",
                },
                duration=300.4,
                check_in_id="1234567890",
                status=MonitorStatus.OK,
            )


def test_crons_task_failure():
    fake_task = MagicMock()
    fake_task.request = {
        "headers": {
            "sentry-monitor-slug": "test123",
            "sentry-monitor-check-in-id": "1234567890",
            "sentry-monitor-start-timestamp-s": 200.1,
            "sentry-monitor-config": {
                "schedule": {
                    "type": "interval",
                    "value": 3,
                    "unit": "day",
                },
                "timezone": "Europe/Vienna",
            },
            "sentry-monitor-some-future-key": "some-future-value",
        },
    }

    with mock.patch(
        "sentry_sdk.integrations.celery.beat.capture_checkin"
    ) as mock_capture_checkin:
        with mock.patch(
            "sentry_sdk.integrations.celery.beat._now_seconds_since_epoch",
            return_value=500.5,
        ):
            crons_task_failure(fake_task)

            mock_capture_checkin.assert_called_once_with(
                monitor_slug="test123",
                monitor_config={
                    "schedule": {
                        "type": "interval",
                        "value": 3,
                        "unit": "day",
                    },
                    "timezone": "Europe/Vienna",
                },
                duration=300.4,
                check_in_id="1234567890",
                status=MonitorStatus.ERROR,
            )


def test_crons_task_retry():
    fake_task = MagicMock()
    fake_task.request = {
        "headers": {
            "sentry-monitor-slug": "test123",
            "sentry-monitor-check-in-id": "1234567890",
            "sentry-monitor-start-timestamp-s": 200.1,
            "sentry-monitor-config": {
                "schedule": {
                    "type": "interval",
                    "value": 3,
                    "unit": "day",
                },
                "timezone": "Europe/Vienna",
            },
            "sentry-monitor-some-future-key": "some-future-value",
        },
    }

    with mock.patch(
        "sentry_sdk.integrations.celery.beat.capture_checkin"
    ) as mock_capture_checkin:
        with mock.patch(
            "sentry_sdk.integrations.celery.beat._now_seconds_since_epoch",
            return_value=500.5,
        ):
            crons_task_retry(fake_task)

            mock_capture_checkin.assert_called_once_with(
                monitor_slug="test123",
                monitor_config={
                    "schedule": {
                        "type": "interval",
                        "value": 3,
                        "unit": "day",
                    },
                    "timezone": "Europe/Vienna",
                },
                duration=300.4,
                check_in_id="1234567890",
                status=MonitorStatus.ERROR,
            )


def test_get_monitor_config_crontab():
    app = MagicMock()
    app.timezone = "Europe/Vienna"

    # schedule with the default timezone
    celery_schedule = crontab(day_of_month="3", hour="12", minute="*/10")

    monitor_config = _get_monitor_config(celery_schedule, app, "foo")
    assert monitor_config == {
        "schedule": {
            "type": "crontab",
            "value": "*/10 12 3 * *",
        },
        "timezone": "UTC",  # the default because `crontab` does not know about the app
    }
    assert "unit" not in monitor_config["schedule"]

    # schedule with the timezone from the app
    celery_schedule = crontab(day_of_month="3", hour="12", minute="*/10", app=app)

    monitor_config = _get_monitor_config(celery_schedule, app, "foo")
    assert monitor_config == {
        "schedule": {
            "type": "crontab",
            "value": "*/10 12 3 * *",
        },
        "timezone": "Europe/Vienna",  # the timezone from the app
    }

    # schedule without a timezone, the celery integration will read the config from the app
    celery_schedule = crontab(day_of_month="3", hour="12", minute="*/10")
    celery_schedule.tz = None

    monitor_config = _get_monitor_config(celery_schedule, app, "foo")
    assert monitor_config == {
        "schedule": {
            "type": "crontab",
            "value": "*/10 12 3 * *",
        },
        "timezone": "Europe/Vienna",  # the timezone from the app
    }

    # schedule without a timezone, and an app without timezone, the celery integration will fall back to UTC
    app = MagicMock()
    app.timezone = None

    celery_schedule = crontab(day_of_month="3", hour="12", minute="*/10")
    celery_schedule.tz = None
    monitor_config = _get_monitor_config(celery_schedule, app, "foo")
    assert monitor_config == {
        "schedule": {
            "type": "crontab",
            "value": "*/10 12 3 * *",
        },
        "timezone": "UTC",  # default timezone from celery integration
    }


def test_get_monitor_config_seconds():
    app = MagicMock()
    app.timezone = "Europe/Vienna"

    celery_schedule = schedule(run_every=3)  # seconds

    with mock.patch("sentry_sdk.integrations.logger.warning") as mock_logger_warning:
        monitor_config = _get_monitor_config(celery_schedule, app, "foo")
        mock_logger_warning.assert_called_with(
            "Intervals shorter than one minute are not supported by Sentry Crons. Monitor '%s' has an interval of %s seconds. Use the `exclude_beat_tasks` option in the celery integration to exclude it.",
            "foo",
            3,
        )
        assert monitor_config == {}


def test_get_monitor_config_minutes():
    app = MagicMock()
    app.timezone = "Europe/Vienna"

    # schedule with the default timezone
    celery_schedule = schedule(run_every=60)  # seconds

    monitor_config = _get_monitor_config(celery_schedule, app, "foo")
    assert monitor_config == {
        "schedule": {
            "type": "interval",
            "value": 1,
            "unit": "minute",
        },
        "timezone": "UTC",
    }

    # schedule with the timezone from the app
    celery_schedule = schedule(run_every=60, app=app)  # seconds

    monitor_config = _get_monitor_config(celery_schedule, app, "foo")
    assert monitor_config == {
        "schedule": {
            "type": "interval",
            "value": 1,
            "unit": "minute",
        },
        "timezone": "Europe/Vienna",  # the timezone from the app
    }

    # schedule without a timezone, the celery integration will read the config from the app
    celery_schedule = schedule(run_every=60)  # seconds
    celery_schedule.tz = None

    monitor_config = _get_monitor_config(celery_schedule, app, "foo")
    assert monitor_config == {
        "schedule": {
            "type": "interval",
            "value": 1,
            "unit": "minute",
        },
        "timezone": "Europe/Vienna",  # the timezone from the app
    }

    # schedule without a timezone, and an app without timezone, the celery integration will fall back to UTC
    app = MagicMock()
    app.timezone = None

    celery_schedule = schedule(run_every=60)  # seconds
    celery_schedule.tz = None

    monitor_config = _get_monitor_config(celery_schedule, app, "foo")
    assert monitor_config == {
        "schedule": {
            "type": "interval",
            "value": 1,
            "unit": "minute",
        },
        "timezone": "UTC",  # default timezone from celery integration
    }


def test_get_monitor_config_unknown():
    app = MagicMock()
    app.timezone = "Europe/Vienna"

    unknown_celery_schedule = MagicMock()
    monitor_config = _get_monitor_config(unknown_celery_schedule, app, "foo")
    assert monitor_config == {}


def test_get_monitor_config_default_timezone():
    app = MagicMock()
    app.timezone = None

    celery_schedule = crontab(day_of_month="3", hour="12", minute="*/10")

    monitor_config = _get_monitor_config(celery_schedule, app, "dummy_monitor_name")

    assert monitor_config["timezone"] == "UTC"


def test_get_monitor_config_timezone_in_app_conf():
    app = MagicMock()
    app.timezone = "Asia/Karachi"

    celery_schedule = crontab(day_of_month="3", hour="12", minute="*/10")
    celery_schedule.tz = None

    monitor_config = _get_monitor_config(celery_schedule, app, "dummy_monitor_name")

    assert monitor_config["timezone"] == "Asia/Karachi"


def test_get_monitor_config_timezone_in_celery_schedule():
    app = MagicMock()
    app.timezone = "Asia/Karachi"

    panama_tz = datetime.timezone(datetime.timedelta(hours=-5), name="America/Panama")

    celery_schedule = crontab(day_of_month="3", hour="12", minute="*/10")
    celery_schedule.tz = panama_tz

    monitor_config = _get_monitor_config(celery_schedule, app, "dummy_monitor_name")

    assert monitor_config["timezone"] == str(panama_tz)


@pytest.mark.parametrize(
    "task_name,exclude_beat_tasks,task_in_excluded_beat_tasks",
    [
        ["some_task_name", ["xxx", "some_task.*"], True],
        ["some_task_name", ["xxx", "some_other_task.*"], False],
    ],
)
def test_exclude_beat_tasks_option(
    task_name, exclude_beat_tasks, task_in_excluded_beat_tasks
):
    """
    Test excluding Celery Beat tasks from automatic instrumentation.
    """
    fake_apply_entry = MagicMock()

    fake_scheduler = MagicMock()
    fake_scheduler.apply_entry = fake_apply_entry

    fake_integration = MagicMock()
    fake_integration.exclude_beat_tasks = exclude_beat_tasks

    fake_client = MagicMock()
    fake_client.get_integration.return_value = fake_integration

    fake_schedule_entry = MagicMock()
    fake_schedule_entry.name = task_name

    fake_get_monitor_config = MagicMock()

    with mock.patch(
        "sentry_sdk.integrations.celery.beat.Scheduler", fake_scheduler
    ) as Scheduler:  # noqa: N806
        with mock.patch(
            "sentry_sdk.integrations.celery.sentry_sdk.get_client",
            return_value=fake_client,
        ):
            with mock.patch(
                "sentry_sdk.integrations.celery.beat._get_monitor_config",
                fake_get_monitor_config,
            ) as _get_monitor_config:
                # Mimic CeleryIntegration patching of Scheduler.apply_entry()
                _patch_beat_apply_entry()
                # Mimic Celery Beat calling a task from the Beat schedule
                Scheduler.apply_entry(fake_scheduler, fake_schedule_entry)

                if task_in_excluded_beat_tasks:
                    # Only the original Scheduler.apply_entry() is called, _get_monitor_config is NOT called.
                    assert fake_apply_entry.call_count == 1
                    _get_monitor_config.assert_not_called()

                else:
                    # The original Scheduler.apply_entry() is called, AND _get_monitor_config is called.
                    assert fake_apply_entry.call_count == 1
                    assert _get_monitor_config.call_count == 1


@pytest.mark.parametrize(
    "task_name,exclude_beat_tasks,task_in_excluded_beat_tasks",
    [
        ["some_task_name", ["xxx", "some_task.*"], True],
        ["some_task_name", ["xxx", "some_other_task.*"], False],
    ],
)
def test_exclude_redbeat_tasks_option(
    task_name, exclude_beat_tasks, task_in_excluded_beat_tasks
):
    """
    Test excluding Celery RedBeat tasks from automatic instrumentation.
    """
    fake_apply_async = MagicMock()

    fake_redbeat_scheduler = MagicMock()
    fake_redbeat_scheduler.apply_async = fake_apply_async

    fake_integration = MagicMock()
    fake_integration.exclude_beat_tasks = exclude_beat_tasks

    fake_client = MagicMock()
    fake_client.get_integration.return_value = fake_integration

    fake_schedule_entry = MagicMock()
    fake_schedule_entry.name = task_name

    fake_get_monitor_config = MagicMock()

    with mock.patch(
        "sentry_sdk.integrations.celery.beat.RedBeatScheduler", fake_redbeat_scheduler
    ) as RedBeatScheduler:  # noqa: N806
        with mock.patch(
            "sentry_sdk.integrations.celery.sentry_sdk.get_client",
            return_value=fake_client,
        ):
            with mock.patch(
                "sentry_sdk.integrations.celery.beat._get_monitor_config",
                fake_get_monitor_config,
            ) as _get_monitor_config:
                # Mimic CeleryIntegration patching of RedBeatScheduler.apply_async()
                _patch_redbeat_apply_async()
                # Mimic Celery RedBeat calling a task from the RedBeat schedule
                RedBeatScheduler.apply_async(
                    fake_redbeat_scheduler, fake_schedule_entry
                )

                if task_in_excluded_beat_tasks:
                    # Only the original RedBeatScheduler.maybe_due() is called, _get_monitor_config is NOT called.
                    assert fake_apply_async.call_count == 1
                    _get_monitor_config.assert_not_called()

                else:
                    # The original RedBeatScheduler.maybe_due() is called, AND _get_monitor_config is called.
                    assert fake_apply_async.call_count == 1
                    assert _get_monitor_config.call_count == 1
